
#        =======================================================
#         ASCII 2 BINARY Table Translation Routines (Ascii2Bin)
#        =======================================================

# Routines for the translation of a FITS file ASCII table into a binary one.
# The normal entry point is proc Ascii2Bin:ConvertAsciiTable below.
# These were/are used for FMOS whose fibre table is in ASCII.

#******************************************************************************
#
# For given FITS object locate the HDU of ASCII table with given name.
#
# Arguments
#   fob       - FITS object
#   tablename - desired table
#
# Return HDU index, or -1 if no such table
#
proc Ascii2Bin:FindAsciiTableHDU {fob tablename} {

    set max_hdu 100

    # Save current HDU so can restore when finished
    set chdu [$fob info chdu]
	
    # Move through all HDU's
    # There is no clean way of determining how many HDUs there are so keep
    # moving until catched error arises which will be "try to pass end of
    # file error". For each HDU check keyword EXTNAME (if exists) for required
    # table name. Also check table is ASCII.
    set found_flag 0; # 0=false
    for {set idx 1} {$idx<=$max_hdu} {incr idx 1} {
	
	# Try to move to next HDU
        # If error then assume have come to file end and quit loop
	set err_flag [catch {$fob move $idx}]
	if {$err_flag} { break; # Have come to end of HDUs }
	    
        # If this type is not ASCII table then skip
	set type [$fob info hdutype]
	if {$type!="ASCII Table"} { continue }
	
	# Get keyword definition list for EXTNAME
	# If error assume keyword does not exist and skip this HDU
        set err_flag [catch {set kywd_list [$fob get keyword EXTNAME]}]
	if {$err_flag} { continue }
	    
        # Returned list is within list (so strip)
        # required name is 2nd item in stripped list
	set kywd_list [lindex $kywd_list 0]
	set  name [lindex $kywd_list 1]
		
	# Check name is table name being searched for
        # If so then found and quit loop
	if {$name=="'${tablename}'"} {
	    set found_flag 1;#1=true
	    break
	}
	}; # end for
	
    # Return current HDU
    $fob move $chdu
	
    # If found return HDU, otherwise return -1
    if {!$found_flag} { return -1 }
    return $idx
	
};# end Ascii2Bin:FindAsciiTableHDU
		
	
#******************************************************************************
#
# Goes through header of current HDU and retrieves a list of header lines that
# are not related to the table definition and are not comments (so we assume
# are potentially important extra info.
#
# Arguments
#   fob       - FITS object
#
# Return list of ???
#
proc Ascii2Bin:GetExtraTableHeaders {fob} {

    set ignore_list {XTENSION BITPIX NAXIS NAXIS1 NAXIS2\
	                 PCOUNT GCOUNT TFIELDS}	
    set ignore_list2 {TTYPE TBCOL TFORM TUNIT} 

    # Dump whole header data into array headerA()
    set data [$fob dump]
    set use_list ""
    foreach line $data {

	# Find "/ comment"
	set idx [string first / $line]
	if {$idx!=-1} {
	    set comment [string range $line [expr $idx+1] end]
	    set line    [string range $line  0 [expr $idx-1]]
	    set list [split $line =]
	    set name  [lindex $list 0]
	    set value [lindex $list 1]
	    set name  [string trim $name " "]
			
	    if {[lsearch $ignore_list $name]!=-1} {continue}
	    set trim_name [string range $name 0 4]
	    if {[lsearch $ignore_list2 $trim_name]!=-1} {continue}
	    
	    lappend use_list $line
	};# endif
    };# end foreach
	
    return $use_list
};# end Ascii2Bin:GetExtraTableHeaders


#******************************************************************************
#
# For the given fits object assume the current HDU is an ASCII table and get
# the list of FORTRAN style data formats for each column.
#
# Arguments
#   fob       - FITS object
#
# Return list of formats
#
proc Ascii2Bin:GetAsciiTableFormats {fobr} {

    set format_list ""
    set col_list [$fobr info column]
    foreach column $col_list {
	set column_def [$fobr info column $column]
	set column_def [lindex $column_def 0]
	set format [lindex $column_def 1]
	lappend format_list $format
    }
	
    return $format_list
};# End Ascii2Bin:GetAsciiTableFormats


#******************************************************************************
#
# For a given list of ASCII formats from a FITS ASCII table translate into a
# list of binary formats.
#
# Arguments
#   ascii_format_list - ???
#
# Return list of binary formats
#
proc Ascii2Bin:TranslateFormatList {ascii_format_list} {

    set bin_format_list ""

    foreach ascii_format $ascii_format_list {
	set bin_format [Ascii2Bin:TranslateFormat $ascii_format]
	lappend bin_format_list $bin_format
    }
	
    return $bin_format_list
};# end Ascii2Bin:TranslateFormatList


#******************************************************************************
#
# For a given ASCII formats from a fits ASCII table translate into a binary
# formats.
#
# Arguments
#   ascii_format - ???
#
# Return binary format
#
proc Ascii2Bin:TranslateFormat {ascii_format} {

    set code [string range $ascii_format 0 0]
    set num  [string range $ascii_format 1 end]
	
    switch $code {
	"I"     {set bin_format "1I"}
	"E"     {set bin_format "1D"}
	"D"     {set bin_format "1D"}
	"F"     {set bin_format "1D"}
	"A"     {set bin_format "${num}A"}
	default {error "unrecognised ascii format $ascii_format"}
    }
};# end Ascii2Bin:TranslateFormat


#******************************************************************************
#
# For a given list of names and a name map list 
#
# Arguments
#   old_names    - ???
#   namemap_list - ???
#
# Return list of new names
#
proc Ascii2Bin:TranslateNames {old_names namemap_list} {

    if {$namemap_list==""} {return $old_names}

    # Parse namemap_list into two lists from_list and corresponding to_list 
    set from_name_list ""; set to_name_list ""
    foreach remap $namemap_list {
	set from_name [lindex $remap 0]
	set to_name   [lindex $remap 1]
	lappend  from_name_list $from_name
	lappend  to_name_list   $to_name
    };# end foreach

    set new_names ""
    foreach old_name $old_names {

	# Check if this name is in list to be renamed, if so use mapped to name
	set idx [lsearch $from_name_list $old_name]
	if {$idx!=-1} {
	    set new_name [lindex $to_name_list $idx]
	} else {
	    set new_name $old_name
	};# end if
	lappend new_names $new_name
    };# end foreach
		
    return $new_names
}; # end Ascii2Bin:TranslateNames


#******************************************************************************
#
# Translate the ASCII table, tablename, in reference fits object, fobr, into
# a binary table of same name and insert into the fits object fobn.
# Note fobr and fobn can be the same.
#
# Arguments
#   fobr         - FITS object
#   fobn         - ???
#   tablename_r  - ???
#   tablename_n  - ???
#   namemap_list - optional, a list of name remaps, i.e. what column names in
#                  the ASCII table should be named in the binary
#
# Return ???
#
proc Ascii2Bin:TranslateAsciiTable {fobr fobn tablename_r tablename_n {namemap_list ""}} {

    # Locate ASCII Table HDU in ref fits object and move to it
    set hdu_idx [Ascii2Bin:FindAsciiTableHDU $fobr $tablename_r]
    if {$hdu_idx==-1} {return -1;# -1=Failed to find table}

    # Make this current HDU
    $fobr move $hdu_idx
	
    # Get table data information
    # Get data in column list format
    set t_data    [$fobr get table -c -noformat];
    set n_cols    [$fobr info ncols]
    set n_rows    [$fobr info nrows]
    set col_names [$fobr info column]

    # Get ASCII column data formats and translate into binary formats
    set ascii_formats  [Ascii2Bin:GetAsciiTableFormats $fobr]
    set bin_formats    [Ascii2Bin:TranslateFormatList $ascii_formats]
    set bin_colnames   [Ascii2Bin:TranslateNames $col_names $namemap_list]
	
    # Get all table header data not related to table data/format
    # Assume this is required in binary table as well
    set extra_headers [Ascii2Bin:GetExtraTableHeaders $fobr]

    # Add binary table placeholder (with null data)
    $fobn insert table $n_rows $bin_colnames $bin_formats "" $tablename_n

    # Go through each table column and populate data into binary table
    set n_cols [llength $col_names]
    for {set idx 0} {$idx<$n_cols} {incr idx 1} {
	set col_data [lindex $t_data $idx]
	set col_name [lindex $bin_colnames $idx]
	$fobn put table $col_name 1 - $col_data
    }

    # Add extra header data from ASCII data into binary table header
    foreach kywd_record $extra_headers {
	$fobn put keyword  $kywd_record
    }

    # Set new tablename hardway as easy way above doesn't appear to work
    $fobn put keyword "EXTNAME=$tablename_n"

};# End Ascii2Bin:TranslateAsciiTable

#******************************************************************************
#
# Translate the ASCII table, tablename, in the fits object, fobr, into a
# binary table.
#
# Arguments
#   fob           - FITS object
#   tablename     - ???
#   tablename_new - ???
#   namemap_list  - optional, a list of name remaps, i.e. what column names in
#                   the ASCII table should be named in the binary
#
# Return ???
#
proc Ascii2Bin:ConvertAsciiTable {fob tablename tablename_new {namemap_list ""}} {

    # Find ASCII Table
    set hdu_idx [Ascii2Bin:FindAsciiTableHDU $fob $tablename]
    if {$hdu_idx==-1} {return -1;# -1=Failed to find table}

    # Create binary translation of table
    Ascii2Bin:TranslateAsciiTable $fob $fob $tablename $tablename_new\
			   $namemap_list
    #Ascii2Bin:TranslateAsciiTable_old $fob $fob $tablename
	
    # Delete ASCII table
    $fob move $hdu_idx
    $fob delete chdu
	
    return 0; #0=OK
	
};# End Ascii2Bin:ConvertAsciiTable 

