#******************************************************************************
#
# Tcl procedures to manipulate FITS files.
#
# "FitsOps" procedures use filenames as arguments, while "FitsFobOps"
# procedures use fobs (FITS object pointers) as arguments.
#
# Notice the procedures here and those in drcontrol/drcFiles.tcl have similar
# (but distinct) functionality.
#
#******************************************************************************

#******************************************************************************
#
# Check the specified file exists and is a valid FITS file, i.e. that it can
# be opened to be read. Originally we just applied a catch to the attempt
# at opening the fits file as a fits object, i.e. catch {fits open }, but
# this caused a memory leak, so now we just attempt to open the file as
# binary and read the main keyword headers to check that that part makes
# sense.
#
# Arguments
#   filename - FITS filename
#
# Returns 0 if invalid (0= false)
#         1 if valid   (1= true)
#
proc FitsOps:ValidFitsFile {filename {time_stamp_check 1}} {

   return [ChkFitsFileMainCmpts $filename]

}


# *****************************************************************************
#
# Private routine used only by FitsOps:ValidFitsFile. Attempts to open
# the given filename in binary, read the main header keywords to check if
# they make sense. Returns 1 if OK and 0 if not.
#
proc ChkFitsFileMainCmpts {fname} {

    set MAX_NO_OF_KEYWORDS 500

    set valid_flag 0
    
    # Check file exists
    if {![file exists $fname]} {return 0}

    # Check file is readable
    if {![file readable $fname]} {return 0}
    
    # Get the file size in bytes
    set fsize [file size $fname]
    
    # Check that there is at least 80 bytes to read
    if {$fsize<80} {return 0}


    # Check that we can actually open this file
    set err_flag [catch {set fid [open $fname r ]}]
    if {$err_flag} {return 0}
   
    # Now attempt to read the main header  information
    fconfigure $fid -translation binary
   
    set end_found 0
    for {set loop 1} {$loop<=$MAX_NO_OF_KEYWORDS} {incr loop 1} {

       # Attempt to read 80 bytes
       set err_flag [catch {set str [read $fid 80]}]
       if {$err_flag} {close $fid; return 0}
       
       # Check for the main keywords "END" BITPIX and NAXIS?
       
       if {[string range $str 0 2]=="END"} {
           set end_found 1
           break
       }
       
       if {[string range $str 0 4]=="NAXIS" || \
           [string range $str 0 5]=="BITPIX"} {
           set lst [split $str =/]      
           eval "set [lindex $lst 0] [lindex $lst 1]"
       }
       
    };# End for
    
    close $fid
    
    # Check if the END key word was found
    if {!$end_found} {return 0}

    # Check if the NAXIS keyword was found
    if {![info exists NAXIS]} {return 0}

    # Check if the NAXIS1, NAXIS2, etc keywords were found
    for {set idx 1} {$idx<=$NAXIS} {incr idx 1} {
        if {![info exists NAXIS${idx}]} {return 0}
    }
    
    # Check if the bitpix keyword was found
    if {![info exists BITPIX]} {return 0}
    
    # Calculate the main image size in bytes
    set prod 1
    for {set idx 1} {$idx<=$NAXIS} {incr idx 1} {
        set n [set NAXIS${idx}]
        set prod [expr $prod * $n]
    }
    set  img_size [expr $prod * abs($BITPIX)/8]

    # Check that the file size is bigger than the main image size
    if {$fsize<$img_size} {return 0}

    # Everything has checked out okay so return 1    
    return 1
        

};# End ChkFitsFileMainCmpts

#******************************************************************************
#
# Provide basic statistics about the main image of the given FITS file.
#
# Arguments
#   filename - FITS filename
#
# Optional Args:
#   -np   - Do not print to standard out in nice format
# Returns:
#     list: min max mean sd median
#
proc FitsOps:StatImage {fitsfile args} {

    set fob [fits open $fitsfile 0]
    set imgdim   [$fob info imgdim]
    set dim1     [lindex $imgdim 0]
    set dim2     [lindex $imgdim 1]

    # load image data into memory and get address pointer
    set ptr [$fob load iblock -- 1 $dim1 1 $dim2]

    # Calculate each statistic using the vexpr command
    set min   [lindex [vexpr -use FitsOps:getPtrData "min(ptr)"] 0]
    set max   [lindex [vexpr -use FitsOps:getPtrData "max(ptr)"] 0]
    set mean  [lindex [vexpr -use FitsOps:getPtrData "average(ptr)"] 0]
    set stdev [lindex [vexpr -use FitsOps:getPtrData "stddev(ptr)"] 0]
    set medi  [lindex [vexpr -use FitsOps:getPtrData "median(ptr)"] 0]

    # Free the pointer and fob handle
    fits free [lindex $ptr 0]
    $fob close

    # If not -np option (no print) then print the results
    if {[lsearch $args "-np"]==-1} {
	puts "\nStats for $fitsfile:"
	puts "\tRange = \[$min , $max\]"
	puts "\tMean  = $mean \t SD = $stdev"
	puts "\tMedian= $medi\n"
    }

    # return the statlist
    return [list $min $max $mean $stdev $medi]
}

#******************************************************************************
#
# Provide the centile values of the image of the given FITS file.
#
# Arguments
#   filename - FITS filename
#
# Optional Args:
#   -np   - Do not print to standard out in nice format
#   -cent [centval]  Return the specified (real) centile
#
# Returns:
#     list: of (integer) centile values unless -cent option is used
#
proc FitsOps:StatImage2 {fitsfile args} {

    # OPen fits file
    set fob [fits open $fitsfile 0]
        
    # Get the image data into a list
    set inp_l [$fob get image]
    
    # Free the fob handle
    $fob close

    # Sort the list
    set sort_l [lsort -real $inp_l]
    set n [llength $sort_l]
    
    # Calc the median
    if {[expr $n%2]==0} {
	    # Even no in list
	    set idx1 [expr $n/2]
	    set idx2 [expr $idx1+1]
	    set val1 [lindex $sort_l $idx1]
	    set val2 [lindex $sort_l $idx2]
	    set median [expr 0.5*($val1+$val2)]
    } else {
	    # Odd no in list
	    set idx [expr ($n-1)/2]
	    set median [lindex $sort_l $idx]
    }

    # Get the min and max
    set min [lindex $sort_l 0]
    set max [lindex $sort_l end]


    # Get the centiles
    set cent_list $min
    for {set centile 1} {$centile<=100} {incr centile 1} {
	    set idx [expr int($centile*$n/100)-1]
	    set centileA($centile) [lindex $sort_l $idx]
	    lappend cent_list $centileA($centile)
    }
     
    # If not -np option (no print) then print the results
    if {[lsearch $args "-np"]==-1} {
        puts "min=$min max=$max median=$median"    
        for {set centile 1} {$centile<10} {incr centile 1} {
           puts "Centile ${centile} %\t $centileA($centile)"
        }
        for {set centile 10} {$centile<=90} {incr centile 10} {
	   puts "Centile ${centile} %\t $centileA($centile)"
        }
        for {set centile 91} {$centile<=100} {incr centile 1} {
	   puts "Centile ${centile} %\t $centileA($centile)"
        }
    }
    
    # If -cent [centval] option then return required value
    if {[set arg_idx [lsearch $args "-cent"]]!=-1} {
       incr arg_idx 1
       set cent [lindex $args $arg_idx]
       if {$cent<0.0}   {return $min}
       if {$cent>100.0} {return $max}
       set idx [expr int($cent*$n/100)-1]
       return [lindex $sort_l $idx]
    }
       
    # Return the centile list   
    return $cent_list
}

#******************************************************************************
#
# Creates a new FITS file based on the given FITS file with the image data
# scaled by supplied factor.
#
# Arguments
#   inpfitsfile         - Intput FITS filename 
#   scale_factor        - Factor to scale the image by
#   outfitsfilefilename - Output FITS filename
#
proc FitsOps:Scale {inpfitsfile scale_factor outfitsfile} {

    if {[file exists $outfitsfile]} {
	file delete $outfitsfile
    } 

    set fob1 [fits open $inpfitsfile 0];# Read only
    set fob2 [fits open $outfitsfile 2];# Write only

    # Get the input image dimensions
    set imgdim   [$fob1 info imgdim]
    set dim1     [lindex $imgdim 0]
    set dim2     [lindex $imgdim 1]

    # load image data into memory and get address pointer
    set img [$fob1 get image]

    # Calculate each statistic using the vexpr command
    set outimg [vexpr "$scale_factor * img"]

    # put difference into output file
    $fob2 insert image -32 2 $imgdim
    $fob2 put image 1 $outimg

    # Free the  fob handles
    $fob1 close
    $fob2 close

    # release image lists
    unset img
    unset outimg

     # copy input file header info into output file
    FitsOps:CopyHeader $inpfitsfile $outfitsfile
}

#******************************************************************************
#
# Creates a new FITS file based on the given FITS file with the image data
# shifted by supplied value.
#
# Arguments
#   inpfitsfile         - Intput FITS filename 
#   shift_factor        - Value to shift the image by
#   outfitsfilefilename - Output FITS filename
#
proc FitsOps:Shift {inpfitsfile shift_factor outfitsfile} {

    if {[file exists $outfitsfile]} {
	file delete $outfitsfile
    } 

    set fob1 [fits open $inpfitsfile 0];# Read only
    set fob2 [fits open $outfitsfile 2];# Write only

    # Get the input image dimensions
    set imgdim   [$fob1 info imgdim]
    set dim1     [lindex $imgdim 0]
    set dim2     [lindex $imgdim 1]

    # load image data into memory and get address pointer
    set img [$fob1 get image]

    # Calculate each statistic using the vexpr command
    set outimg [vexpr "$shift_factor + img"]

    # put difference into output file
    $fob2 insert image -32 2 $imgdim
    $fob2 put image 1 $outimg

    # Free the  fob handles
    $fob1 close
    $fob2 close

    # release image lists
    unset img
    unset outimg

     # copy input file header info into output file
    FitsOps:CopyHeader $inpfitsfile $outfitsfile
}
#******************************************************************************
#
# Creates a new FITS file based on the given FITS file with the image data
# logically operated on by specified operation string.
#
# Arguments
#   inpfitsfile         - Intput FITS filename 
#   logical_op          - Logical operation (eg ">7203", <0.0, "!=66363")
#   outfitsfilefilename - Output FITS filename
#
proc FitsOps:LogicalOp {inpfitsfile logical_op outfitsfile} {

    if {[file exists $outfitsfile]} {
	file delete $outfitsfile
    } 

    set fob1 [fits open $inpfitsfile 0];# Read only
    set fob2 [fits open $outfitsfile 2];# Write only

    # Get the input image dimensions
    set imgdim   [$fob1 info imgdim]
    set dim1     [lindex $imgdim 0]
    set dim2     [lindex $imgdim 1]

    # load image data into memory and get address pointer
    set img [$fob1 get image]

    # Calculate logical outcome the vexpr command
    set outimg [vexpr "img $logical_op"]

    # put difference into output file
    $fob2 insert image -32 2 $imgdim
    $fob2 put image 1 $outimg

    # Free the  fob handles
    $fob1 close
    $fob2 close

    # release image lists
    unset img
    unset outimg

     # copy input file header info into output file
    FitsOps:CopyHeader $inpfitsfile $outfitsfile
}

#******************************************************************************
#
# Creates a new FITS file based on the given FITS file list with the image data
# logically "AND" ed
#
# Arguments
#   inpfitsfile         - Intput FITS filename 
#   logical_op          - Logical operation (eg ">7203", <0.0, "!=66363")
#   outfitsfilefilename - Output FITS filename
#
proc FitsOps:AND {inpfitsfile0 inpfitsfile1 outfitsfile} {

    if {[file exists $outfitsfile]} {
	file delete $outfitsfile
    } 

    set fob0 [fits open $inpfitsfile0 0];# Read only
    set fob1 [fits open $inpfitsfile1 0];# Read only
    set fob2 [fits open $outfitsfile 2];# Write only

    # Get the input image dimensions
    set imgdim   [$fob0 info imgdim]
    set dim1     [lindex $imgdim 0]
    set dim2     [lindex $imgdim 1]

    # load image data into memory and get address pointer
    set img0 [$fob0 get image]
    set img1 [$fob1 get image]
    set size [llength $img0]

    # Calculate logical outcome the vexpr command
    #set outimg [vexpr "img0 .OR. img1"]
    # The above (commented out line) results in a hang so we do it the long way.
    set outimg ""
    for {set i 0} {$i<$size } {incr i 1} {
        set bool1 [lindex $img0 $i]
	set bool2 [lindex $img1 $i]
	if {$bool1==0 || $bool2==0} {set bool3 0} else {set bool3 1}
    	lappend outimg $bool3
    }

    # put difference into output file
    $fob2 insert image -32 2 $imgdim
    $fob2 put image 1 $outimg

    # Free the  fob handles
    $fob0 close
    $fob1 close
    $fob2 close

    # release image lists
    unset img0
    unset img1
    unset outimg

     # copy input file header info into output file
    FitsOps:CopyHeader $inpfitsfile0 $outfitsfile
}

#******************************************************************************
#
# Dumps the pixel values of a specified row of the main image of the given
# FITS file into a specified text file.
#
# Arguments
#   inpfitsfile   - Intput FITS filename 
#   row_no        - Row no to extract from
#   outfile       - Output text filename
#
proc FitsOps:DumpRow {inpfitsfile row_no outfile} {

    # Open output text file
    if {[file exists $outfile]} {
	file delete $outfile
    }

    set fid [open $outfile w]

    # Open fits file
    set fob [fits open $inpfitsfile 0];# Read only

    # Load image into pointer
    set myRow [$fob load irows $row_no $row_no]

    # Close the fits file
    $fob close

    # Convert pointer to tcl list
    set lst [eval "ptr2lst $myRow"]

    # Output tcl list to text file
    set idx 1
    foreach itm $lst {
	puts $fid "$idx $itm"
	incr idx 1
    }

    # Close text file
    close $fid

    # Free pointer allocation and data list
    fits free [lindex $myRow 0]
    unset lst
}

#******************************************************************************
#
# Create a new FITS file by subtracting the image data from two given FITS
# files.
#
# Arguments
#   minuefile  - minuend FITS filename
#   subtrafile - subtrahend FITS filename
#   outfile    - FITS filename to be made from difference
#
# Note this procedure is inefficient as the operations are performed in tcl.
#
# Returns ???
#
proc FitsOps:Subtract {minuefile subtrafile outfile} {

    # open minuend & subtrahend files for read only
    set fits_obj1 [fits open $minuefile 0]
    set fits_obj2 [fits open $subtrafile 0]
	
    # check their image dimensions are compatible
    set imgdim1 [$fits_obj1 info imgdim]
    set imgdim2 [$fits_obj2 info imgdim]
    if {$imgdim1!=$imgdim2} {
	FitsOps:Error "Images in $minuefile and $subtrafile are incompatible"
	$fits_obj1 close
	$fits_obj2 close
	return
    }
	
    # open output file for writing
    set fits_obj3 [fits open $outfile 2]  ;# write only

    # Create image list data that is a subtraction from obj1 and obj2
    set minuend [$fits_obj1 get image]
    set subtrahend [$fits_obj2 get image]
    set diff [vexpr "minuend - subtrahend"]
	
    # put difference into output file
    $fits_obj3 insert image -32 2 $imgdim1
    $fits_obj3 put image 1 $diff
    
    # close file objects
    $fits_obj1 close
    $fits_obj2 close
    $fits_obj3 close
	
    # release image lists
    unset minuend
    unset subtrahend
    unset diff

    # copy minuend file header info into output file
    FitsOps:CopyHeader $minuefile $outfile
}

#******************************************************************************
#
# Create a new FITS file based on the given FITS file with the image data
# transposed, i.e. Anew(i,j)=Aold(j,i)
#
# Note this procedure is inefficient as the operations are performed in tcl.
#
# Arguments
#   ref_filename - reference FITS filename
#   new_filename - output FITS file name
#
# Returns ???
#
proc FitsOps:Transpose {ref_filename new_filename} {

    # open reffilename as FITS object
    set fits_obj [fits open $ref_filename 0]  ;# read only
    
    set imgdim [$fits_obj info imgdim]
    set dim1 [lindex $imgdim 0]
    set dim2 [lindex $imgdim 1]
	
    # Make new file copy of FITS file and open as FITS object
    $fits_obj copy $new_filename
    set new_fits_obj [fits open $new_filename 2]  ;# write only

    # load FITS image from reffile into Tcl array (A)
    $fits_obj load iblock A 1 $dim1 1 $dim2
	
    # create imagelist defined as data array transposed
    set list ""
    for {set i 0} {$i<$dim1} {incr i 1} {
	for {set j 0} {$j<$dim2} {incr j 1} {
	    lappend list $A($i,$j)
	}
    }
	
    # create image placeholder in new fist object
    #   and place image list data into it
    set new_imgdim [list $dim2 $dim1]
    $new_fits_obj insert image -32 2 $new_imgdim
    $new_fits_obj put image 1 $list
	
    # close and tidy up
    $fits_obj close
    $new_fits_obj close
	
    # release image list
    unset list
	
    # copy minuend file header info into output file
    FitsOps:CopyHeader $ref_filename $new_filename
}

#******************************************************************************
#
# Create a new FITS file based on the first of the supplied set with image
# data equal to the average of the supplied FITS images.
#
# Arguments
#   file_list   - list of FITS files to be averaged
#   outfilename - output FITS filename
#
# Returns ???
#
proc FitsOps:Average {file_list outfile} {

    set n [llength $file_list]
    
    set ptr1 [FitsOps:Sum2Ptr $file_list]
	
    # Use vexpr to divide by n and free up unwanted data from memory	
    set ptr [vexpr -ptr -use FitsOps:getPtrData "ptr1/$n"]
    fits free [lindex $ptr1 0]

    # now convert memory data into Tcl list
    set new_list [eval ptr2lst $ptr]
    fits free [lindex $ptr 0]

    # create new FITS file (based on first supplied FITS file)
    # and dump the new_list into it
    set file [lindex $file_list 0]
    set fits_obj [fits open $file]
    set imgdim [$fits_obj info imgdim]
    $fits_obj close
    #$fits_obj copy $outfile

    set new_fits_obj [fits open $outfile 2]  ;# write only
    $new_fits_obj insert image -32 2 $imgdim
    $new_fits_obj put image 1 $new_list
	
    # tidy up
    $new_fits_obj close
    unset new_list

    # copy headers across into new FITS file
    FitsOps:CopyHeader [lindex $file_list 0] $outfile
}

#******************************************************************************
#
# Create a new FITS file based on the first of the supplied set with iamge
# data equal to the subtraction of the average of the first list from the
# average of the second list.
#
# Arguments
#   file_list1  - list of initial FITS files to be averaged
#   file_list2  - list of final FITS files to be averaged
#   outfilename - output FITS filename
#
# Returns ???
#
proc FitsOps:Fowler {file_list1 file_list2 outfile} {

    set n1 [llength $file_list1]
    set n2 [llength $file_list2]
	
    set ptr1 [FitsOps:Sum2Ptr $file_list1]
    set ptr2 [FitsOps:Sum2Ptr $file_list2]
	
    # Use vexpr to divide by n and free up unwanted data from memory
    set ptr [vexpr -ptr -use FitsOps:getPtrData "(ptr2)/$n2-(ptr1)/$n1"]
    fits free [lindex $ptr1 0]
    fits free [lindex $ptr2 0]

    # Now convert the memory data into a tcl list
    set new_list [eval ptr2lst $ptr]
    fits free [lindex $ptr 0]

    # Create a new FITS file (based on the first supplied FITS file)
    # and dump the new_list into it
    set file [lindex $file_list1 0]
    set fits_obj [fits open $file]
    set imgdim [$fits_obj info imgdim]
    $fits_obj close

    # Create a new FITS file (based on the first supplied FITS file)
    # and dump the new_list into it
    set new_fits_obj [fits open $outfile 2]  ;# write only
    $new_fits_obj insert image -32 2 $imgdim
    $new_fits_obj put image 1 $new_list
    $new_fits_obj close
    
    unset new_list
    
    # Copy the header acrros into the new FITS file
    FitsOps:CopyHeader [lindex $file_list1 0] $outfile
}

#******************************************************************************
#
# ???
#
# Returns ???
#
proc FitsOps:Ramp {} {

    puts "Application under development"
}

#******************************************************************************
#
# ???
#
# Arguments
#   filename - 
#   row_idx  -
#
# Returns ???
#
proc FitsOps:SumRow {filename row_idx {range ""}} {

    set fob [fits open $filename 0]  ;# read only
    if {$range==""} {
       set val [FitsFobOps:SumRow $fob $row_idx]
    } else {
       set val [FitsFobOps:SumRow2 $fob $row_idx $range]
    }    
    $fob close
    return $val	
}

#******************************************************************************
#
# ???
#
# Arguments
#   filename - 
#   col_idx  -
#
# Returns ???
#
proc FitsOps:SumCol {filename col_idx} {

    set fob [fits open $filename 0]  ;# read only
    set val [FitsFobOps:SumCol $fob $col_idx]
    $fob close
    return $val	
}

#******************************************************************************
#
# ???
#
# Arguments
#   filename - 
#   ret_type - either "x", "y" or "both"
#
# Returns either the X dimension, Y dimension, or a list containing both.
#
proc FitsOps:ImageSize {filename {ret_type both}} {

    set ret_type [string tolower $ret_type]

    set fob [fits open $filename 0]  ;# read only
    set imgdim [$fob info imgdim]
    set x [lindex $imgdim 0]
    set y [lindex $imgdim 1]
    $fob close

    switch $ret_type {
	"x"	{return $x}
	"nx"	{return $x}
	"cols"	{return $x}
	"ncols"	{return $x}
	"y"	{return $y}
        "ny"    {return $y}
        "rows"  {return $y}
        "nrows" {return $y}
	"both"	{return $imgdim}
    }
}

#******************************************************************************
#
# Extract fibre flux into an ASCII output file.
#
# The intensities along each data row (fibre) are summed to produce a total
# (flux) for each fibre. The total is written along with the fibre number to
# form one line in the output file.
#
# Arguments
#   fitsfile - input reduced 2dfdr data file
#   outfile  - output ASCII file containing fibre flux for "P" type fibres
#
# Returns ""
#
proc FitsOps:FluxEx {fitsfile outfile} {
    
    # open output file for writing
    set fid [open $outfile w]

    # output first line containing mean DEC & RA values from header
    set dec_val [FitsOps:GetKeyWord $fitsfile MEANDEC]
    set ra_val  [FitsOps:GetKeyWord $fitsfile MEANRA]
    puts $fid "MEAN RA=$ra_val	MEAN DEC=$dec_val"
    
    # get Fibre Table Information from input fitsfile
    # (we want to know the fibre types)
    set col_names [FitsOps:TableDump $fitsfile *FIBRES* fibreA]
    set type_idx  [expr [lsearch $col_names "TYPE"] +1]
    
    # open input file and get # of fibres (Y dimension)
    set fob [fits open $fitsfile 0]  ;# read only
    set imgdim [$fob info imgdim]
    set ny [lindex $imgdim 1]

    # sum flux for each fibre and output as output file line
    for {set row_idx 1} {$row_idx<=$ny} {incr row_idx 1}  {
    
        # If this fibre is not a "P" type then skip it.
	set type [string toupper $fibreA($row_idx,$type_idx)]
        if {$type!="P"} {
	    continue
	}
	
	puts $fid "$row_idx	[FitsFobOps:SumRow $fob $row_idx]"
    }

    # close files
    close $fid
    $fob close
}

#******************************************************************************
#
# Return list of extension (HDU) names found in supplied FITS file.
#
# Arguments
#   fitsfile - FITS file name
#
# Returns Tcl list of extension names
#
proc FitsOps:ListExtensions {fitsfile} {
	
    # open file
    set err_flag [catch {
       set fob [fits open $fitsfile 0]  ;# read only
    }]
    if {$err_flag} {puts "e1";catch {$fob close}; return ""}

    # go through each HDU
    set n_hdu [$fob info nhdu]
    set extname_list ""	
    for {set idx 1} {$idx<=$n_hdu} {incr idx 1} {
	
	# make this CHDU
        set err_flag [catch {
 	    $fob move $idx
	}]
        if {$err_flag} {puts "e2"; catch {$fob close}; return ""}
        
	# Get keyword definition list for EXTNAME
	# Some 2dfdr old combined files don't have EXTNAME for primary HDU
	# If error assume keyword does not exist and skip this HDU
	set err_flag [catch {set kywd_list [$fob get keyword EXTNAME]}]
	if {$err_flag && $idx==1} {
	    set kywd_list [list {EXTNAME IMAGE comment}]
	} elseif {$err_flag} {
	    continue
	}
		
	# returned list is within list, so strip
	# name is 2nd item in stripped list
	# name is sometimes list, so strip it too
	set kywd_list [lindex $kywd_list 0]
	set name [lindex $kywd_list 1]
	set name [lindex $name 0]
	set name [string trim $name "'"]
		
	lappend extname_list $name
		
    };# end for
	
    # close file and return extension names list
    $fob close
    return $extname_list
}

#******************************************************************************
#
# Find an extension (HDU) name that matches the supplied pattern
# (e.g. "*.FIBRES") in the supplied FITS file.
#
# Arguments
#   fitsfile - FITS file name containing extension
#   extname  - extension name string match pattern, e.g. "*.FIBRES"
#
# Returns index of extension (HDU), or
#         empty string if extension not found
#
proc FitsOps:FindExtension {fitsfile extname} {

    set ext_list [FitsOps:ListExtensions $fitsfile]
    set idx [lsearch $ext_list $extname]
    if {$idx==-1} {
	return ""
    }
    return [lindex $ext_list $idx]
}

#******************************************************************************
#
# Return specified details of supplied keyword in specified FITS file.
#
# Arguments
#   fitsfile - FITS file name
#              This is either the FITS file or a Tcl list containing the
#              FITS file and an extension (HDU) name, e.g.
#              "6sep00031.fits" or "6sep00031.fits STRUCT.MORE.FIBRES"
#              The primary extension is used if one is not given.
#              Note wildcards may be used for the extension name, e.g.
#              "6sep00031.fits *.FIBRES"
#   keyname  - FITS keyword name to search for
#   option   - Optional argument specifing return value
#              default (empty string) returns the keyword value (with spaces
#                and ' trimmed) or empty string if keyword does not exist
#              "-c" returns the keyword comment value
#              "-e" returns whether the keyword exists (1 or 0)
#
# Returns see "option" argument
#
proc FitsOps:GetKeyWord {fitsfile keyname {option ""}} {
	
    # initialise error flag to false
    set err_flag 0
	
    # open file
    set fits_obj [fits open [lindex $fitsfile 0] 0]  ;# read only
	
    # If extension name is given try to find it and move fob pointer to it
    set extname  [lindex $fitsfile 1]	
    if {$extname!=""} {
	set hdu [FitsFobOps:FindExtensionHDU $fits_obj $extname]
	if {$hdu!=-1} {
	    $fits_obj move $hdu
	} else {
	    set err_flag 1  ;# couldn't find extension
	}
    }
	
    # $fits_obj get keyword command gets all matches and errors if none
    if { ! $err_flag} {
	set err_flag [catch {$fits_obj get keyword $keyname} result]
    }
    if {$err_flag} {
	# here no keyword match found
	set value "";
	set comment ""
    } else {
	# here keyword match found, check first match is in fact keyname
	set record  [lindex $result 0]
	set key     [lindex $record 0]
	set value   [lindex $record 1]
	set comment [lindex $record 2]
	if {$key!=$keyname} {
	    set err_flag 1
	}
    }

    # close file
    $fits_obj close
	
    # return value based on option argument
    if {$option=="-c"} {
	return $comment
    } elseif {$option=="-e"} {
	return [expr !$err_flag]
    } else {
	return [string trim $value " '"]
    }	
}

#******************************************************************************
#
# Add a variable number of keywords into specified FITS file.
#
# Arguments
#   filename - FITS file name
#   args     - keywords in "[name] [value]" format
#
# Returns ???
#
proc FitsOps:AddKeyWord {filename args} {

    set fits_obj [fits open $filename]
	
    foreach key_def $args {
	
	set keyword  [lindex $key_def 0]
	set keyvalue [lindex $key_def 1]
	
	$fits_obj put keyword "$keyword = $keyvalue"
    }

    $fits_obj close
}

#******************************************************************************
#
# Rename a vraiable number of keywords within specified FITS file.
#
# Arguments
#   filename - FITS filename
#   args     - keyword changes in "[old_name] [new_name]" format
#
# Returns ???
#
proc FitsOps:RenameKeyWord {filename args} {
	
    set fits_obj [fits open $filename]
	
    foreach key_def $args {
	
	set old_keyname [lindex $key_def 0]
	set new_keyname [lindex $key_def 1]
	
	set res [$fits_obj get keyword $old_keyname]
	set res [lindex $res 0]
	set value   [lindex $res 1]
	set comment [lindex $res 2]
	$fits_obj put keyword "$new_keyname = $value /$comment"
	$fits_obj delete keyword $old_keyname
    }

    $fits_obj close
}

#******************************************************************************
#
# Copies all the header keywords from a ref FITS file and place them into the
# etarhget FITS file.
#
# Arguments
#   reffilename - reference FITS filename
#   outfilename - target FITS filename
#
# Returns ???
#
proc FitsOps:CopyHeader {reffilename outfilename} {

    set tar_fits_obj [fits open $outfilename 1]  ;# read and write
    set ref_fits_obj [fits open $reffilename 0]  ;# read only

    for {set idx 1} {$idx<[$ref_fits_obj info nkwds]} {incr idx 1} {
	set record [$ref_fits_obj get keyword -num $idx]
		
	set keyname [lindex $record 0]
	set value   [lindex $record 1]
	set comment [lindex $record 2]

	$tar_fits_obj put keyword "$keyname = $value /$comment"
	#$tar_fits_obj put keyword $record
    }
	
    $tar_fits_obj close
    $ref_fits_obj close
}
#******************************************************************************
#
# For a specified table extension (HDU) name in a specified FITS file, dump
# the table data into the supplied Tcl array.
#
# Returned Data Array has following form:
#   dataArray(i,j)       = table data of row i and col j
#   dataArray(n_cols)    = number of columns
#   dataArray(n_rows)    = number of rows
#   dataArray(col_names) = list of column headers in order of column
#
# Arguments
#   fitsfile  - FITS file name
#   tablename - table extension (HDU) name (may contain wildcards)
#   arrayname - returned table contents
#
# Returns list of column names found, or empty string if extension not found
#
proc FitsOps:Image2TxtFile {fitsfile imagename filename} {
	
    # Open the fits file, but in the event of an error
    # return some dummy values
    set fob [fits open $fitsfile 0]  ;# read only
   
    # find HDU with extension name that matches table name
    if {[string tolower $imagename]!="primary"} {
       set hdu [FitsFobOps:FindExtensionHDU $fob $imagename]
       if {$hdu==-1} {
	   $fob close
	   return "No such extension $imagename"
       }

       $fob move $hdu
    }
    set imgdim [$fob info imgdim]
    set lst [$fob get image ]
    $fob close
    set fid [open $filename w]
    #fconfigure $fid -translation binary
    #set binlst [binary format f [lindex $lst 0]]
    #puts $fid $binlst
    #puts $fid $lst
    puts $fid $imgdim
    set str [regsub -all NULL $lst 0.0]
    puts $fid $str
    close $fid
    return $imgdim
	
};# End FitsOps:Image2TxtFile

#******************************************************************************
#
# For a specified table extension (HDU) name in a specified FITS file, dump
# the table data into the supplied Tcl array.
#
# Returned Data Array has following form:
#   dataArray(i,j)       = table data of row i and col j
#   dataArray(n_cols)    = number of columns
#   dataArray(n_rows)    = number of rows
#   dataArray(col_names) = list of column headers in order of column
#
# Arguments
#   fitsfile  - FITS file name
#   tablename - table extension (HDU) name (may contain wildcards)
#   arrayname - returned table contents
#
# Returns list of column names found, or empty string if extension not found
#
proc FitsOps:TableDump {fitsfile tablename arrayname} {
	
    upvar $arrayname dataA

    # Open the fits file, but in the event of an error
    # return some dummy values
    set fob [fits open $fitsfile 0]  ;# read only
   
    set err_flag {
        puts "Table Dump error ${err_msg}"
        set col_names "?"
        set dataA(col_names) $col_names
        set dataA(n_cols) 1
        set dataA(n_rows) 1
        set dataA(1,1) "???"
        return $col_names
    }
        
    # find HDU with extension name that matches table name
    set hdu [FitsFobOps:FindExtensionHDU $fob $tablename]
    if {$hdu==-1} {
	$fob close
	return ""
    }
	
    $fob move $hdu
    set t_data    [$fob get table -noformat]; # get data in column list format
    set n_cols    [$fob info ncols]
    set n_rows    [$fob info nrows]
    set col_names [$fob info column]

    set i 1
    foreach row $t_data {
	set j 1
	foreach elem $row {
	    set dataA($i,$j) $elem
	    incr j 1
	}
	incr i 1
    }
	
    $fob close
	
    set dataA(col_names) $col_names
    set dataA(n_cols) $n_cols
    set dataA(n_rows) $n_rows
    return $col_names
};# End FitsOps:TableDump

#******************************************************************************
#
# Set the dimensions, column names and cell values for a specified
# table extension with data defined in supplied data array.
#
# Table extension must exist (wildcard * may be used in name)
#
# Structure of data array must be as follows:
# dataA(ncols) 	= no of columns
# dataA(nrows) 	= no of rows
# dtatA(col_names) 	= list of column names (in column order)
# data(i,j) 	= data value of cell i,j
#
# Arguments:
#   fitsfile  - FITS file name
#   tablename - table extension (HDU) name (may contain wildcards)
#   arrayname - where table contents are returned
#
# Return value:
#	-1 if could not succeed
#        1 if sucessful

proc FitsOps:SetTableData {filename tablename dataAName} {

    upvar $dataAName dataA

    set fob [fits open $filename 1]
    set hdu [FitsFobOps:FindExtensionHDU $fob $tablename]
    if {$hdu==-1} {return -1}
    $fob move $hdu

    set n_cols    $dataA(n_cols)
    set n_rows    $dataA(n_rows)
    set col_names $dataA(col_names)
    
    for {set j 1} {$j<=$n_cols} {incr j 1} {
        set col_data ""
        for {set i 1} {$i<=$n_rows} {incr i 1} {
            lappend col_data $dataA($i,$j)
        }
        set idx [expr $j-1]
        set col_name [lindex $col_names $idx]
        $fob put table $col_name 1 - $col_data

    }

    $fob close
    
    return 1

}; # End FitsOps:SetTableData


#******************************************************************************
#
# Create a standard 2dfdr fibre table with empty entries that can be filled
# with FitsOps:SetTableData. This is useful is translating a non standard
# instrument fits file into a generic one for 2dfdr to process.
#
# Arguments:
#   fitsfile  - FITS file name
#   nrows     - number of rows in table
#
# Return value:
#	-1 if could not succeed
#        1 if sucessful

proc FitsOps:CreateFibreTable {filename nrows} {
    
    set fmt_lst  {80A 1D 1D 1J 1J 1I 1I 1D 1A 1I 1D 1J 80A 10A 1D}
    set name_lst {NAME RA DEC X Y XERR YERR THETA TYPE PIVOT MAGNITUDE \
                  PID COMMENT RETRACTOR WLEN}

    # First define the fib data array that will define the table elements
    set fibDataA(n_rows) $nrows
    set fibDataA(n_cols) [llength $name_lst]
    set fibDataA(col_names) $name_lst
    for {set row 1} {$row<=$nrows} {incr row 1} {
       set fibDataA($row,1)   G00279938
       set fibDataA($row,10)  1
       set fibDataA($row,11)  19.79
       set fibDataA($row,12)  0
       set fibDataA($row,13)  G00279938_IP6_Y2_140.14_01.07
       set fibDataA($row,14)  68
       set fibDataA($row,15)  600.0
       set fibDataA($row,2)   2.44596113695
       set fibDataA($row,3)   0.0186717262633
       set fibDataA($row,4)   -7172
       set fibDataA($row,5)   56737
       set fibDataA($row,6)   0
       set fibDataA($row,7)   4
       set fibDataA($row,8)   6.264288
       set fibDataA($row,9)   P
    }

    
    # Now create the table template using the basic header keywords
    set fob [fits open $filename]
    $fob insert table  $fibDataA(n_rows) "$fibDataA(col_names)" "$fmt_lst"
    $fob put keyword "EXTNAME=FIBRES"
    $fob close
    FitsOps:SetTableData $filename FIBRES fibDataA
        
    return 1

}; # End FitsOps:CreateFibreTable


#******************************************************************************
#
#
# For a specified image in a specified FITS file, return the
# row's axis data as a list.
#
#
proc FitsOps:GetAxis {filename} {

    # Definition taken from tdfio_axis_read_dp.F95
    set nx [FitsOps:ImageSize $filename "nx"]
    set REFV [FitsOps:GetKeyWord $filename "CRVAL1"]
    set REFP [FitsOps:GetKeyWord $filename "CRPIX1"]
    set DELT [FitsOps:GetKeyWord $filename "CDELT1"]
    set OFFSET [expr $REFV-($REFP-1.0)*$DELT]
    
    set axis_lst ""
    for {set i 1} {$i<=$nx} {incr i 1} {
        set axis_value [expr $DELT * ($i-1) + $OFFSET]
        lappend axis_lst $axis_value
    }
    
    return $axis_lst
    
};# End FitsOps:GetAxis 

#******************************************************************************
#
# For a specified row of the main image in a specified FITS file, return the
# row data as a list.
#
# Returns image row data list
#
proc FitsOps:GetImageRow {filename row} {

    set fob [fits open $filename 0]  ;# read only

    set imgdim [$fob info imgdim]
    set nx [lindex $imgdim 0]
    set ny [lindex $imgdim 1]

    set idx [expr $nx*($row-1)+1]

    set l [$fob get image $idx $nx]
	
    $fob close
	
    return $l
    
};# End FitsOps:GetImageRow


#******************************************************************************
#
# Converts an ASCII table in the specified FITS file into a binary one. Note
# makes use of the Ascii2Bin routines.
#
# Arguments
#   filename     - FITS filename containing the table
#   tablename    - name of the ASCII table to convert
#   namemap_list - Optional argument that is a list of any name remaps, i.e.
#                  what an ascii column name should be in the binary table.
#
# Returns ???
#
proc FitsOps:Ascii2BinTable {filename tablename tablename_new \
				 {namemap_list ""}} {

    set fob [fits open $filename 1]  ;# read and write
	
    set flag [Ascii2Bin:ConvertAsciiTable $fob $tablename \
		  $tablename_new $namemap_list]
	
    $fob close
	
    if {$flag==-1} {
	puts "Unable to find table $tablename in file $filename"
    }
}

#******************************************************************************
#
# For a given FITS object find the index of the specified extension (HDU)
# name (which may have regular expression wildcards).
#
# Arguments
#   fob     - pointer to FITS file object
#   extname - extension name to be searched for
#
# Returns HDU number, or
#         -1 if no such extension exists
#
proc FitsFobOps:FindExtensionHDU {fob extname} {
		
    # HDU is not yet found
    set found_flag 0
    
    set search_pattern '${extname}*'

    # get current HDU index to make it current when finished
    set chdu [$fob info chdu]

    # go through each HDU
    set n_hdu [$fob info nhdu]	
    for {set idx 1} {$idx<=$n_hdu} {incr idx 1} {
	
	# set current HDU
	$fob move $idx
	
	# get EXTNAME keyword definition
	# if error assume keyword does not exist and skip this HDU
	set err_flag [catch {set kywd_list [$fob get keyword EXTNAME]}]
	if {$err_flag} {
	    continue
	}
		
	# returned list is within a list (so strip)
	# required name is 2nd item in stripped list
	set kywd_list [lindex $kywd_list 0]
	set name [lindex $kywd_list 1]
		
	# check name is HDU name being searched for
	# If so then found and quit loop
	if {[string match $search_pattern $name]} {
	    set found_flag 1  ;# 1=true
	    break
	}
    };# end for
	
    # restore original HDU index
    $fob move $chdu
	
    # if HDU was found return its index, otherwise return -1
    if {!$found_flag} {
	return -1
    }
    return $idx

};# end  FitsFobOps:FindExtensionHDU

#******************************************************************************
#
# Sums the given data column.
#
# Arguments
#   fob     - pointer to FITS file object
#   col_idx - column index to summed
#
# Returns ???
#
proc FitsFobOps:SumCol {fob col_idx} {

    set myCol [$fob load icols $col_idx $col_idx]
    lindex [vexpr -use FitsOps:getPtrData "sum( (double)myCol)"] 0
}

#******************************************************************************
#
# Sums the given data row.
#
# Arguments
#   fob     - pointer to FITS file object
#   row_idx - row index to be summed
#
# Returns ???
#
proc FitsFobOps:SumRow {fob row_idx} {

    set myRow [$fob load irows $row_idx $row_idx]
    lindex [vexpr -use FitsOps:getPtrData "sum( (double)myRow)"] 0
}

proc FitsFobOps:SumRow2 {fob row_idx {range ""}} {

    set imgdim [$fob info imgdim]
    set nrows [lindex $imgdim 0]
    set ncols [lindex $imgdim 1]
    if {$range==""} {
    	set beg 1
	set end $ncols
    } else {
        set beg [lindex $range 0]
	set end [lindex $range 1]
    }
    set nelems [expr $end-$beg+1]
    set lst [$fob get image [expr ($row_idx-1)*$ncols+$beg] $nelems]
    set sum [vexpr "sum( (double)lst)"]
    return $sum
    
}
# ------------------------------------------------------------------
#                        ==================
#                         Utility Routines
#                        ==================
# ------------------------------------------------------------------

#******************************************************************************
#
# Used with the fitsTcl vexpr command (via -use) so vexpr will work on array
# pointers.
#
# Arguments
#   varName - ???
#
# Returns ???
#
proc FitsOps:getPtrData {varName} {

    upvar $varName myName
    if {[info exists myName]} {
	return [eval list -ptr $myName]
    } else {
	error "Variable $varName does not exists"
    }
}

#******************************************************************************
#
# Summate the image data in the supplied list of FIT files and return the
# pointer for the summated data.
#
# Arguments
#   file_list   - The list of FITS files to be averaged.
#
# Returns ???
#
proc FitsOps:Sum2Ptr {file_list} {

    set n [llength $file_list]
	
    # iterate through all input FITS files
    # for each file load image data into memory, add into accumulator space
    # using vexpr and then free unwamted data from memory
    for {set idx 0} {$idx<$n} {incr idx 1} {
	
	# open file
	set file     [lindex $file_list $idx]		
	set fits_obj [fits open $file 0]  ;# read only
	set imgdim   [$fits_obj info imgdim]
	set dim1     [lindex $imgdim 0]
	set dim2     [lindex $imgdim 1]

	# load image data into memory and get address pointer
	set ptr [$fits_obj load iblock -- 1 $dim1 1 $dim2]
		
	# close file
	$fits_obj close
		
	if {$idx==0} {
	    set ptr1 $ptr; continue
	} else {
	    set ptr2 $ptr
	}
	
	# add image data using vexpr and get pointer address
	# for the resulting data
	set ptr3 [vexpr -ptr -use FitsOps:getPtrData "ptr1+ptr2"]
	
	# free redundant data 
	fits free [lindex $ptr1 0]
	fits free [lindex $ptr2 0]

	set ptr1 $ptr3
    }

    return $ptr1
}; # End FitsOps:Sum2Ptr
