#******************************************************************************
#
# Compares the main data items (headers, extension list, fibre table and image
# values) of the two given FITS files. Outputs the test result information to
# the optional argument "reportfile" which is either a filename, "stdout" (the
# default) or "stderr".
#
# Optional Arguments
#  -diffpercent <percent>
#	Two values are considered different if they differ by this percent
#	or greater.
#  -failpercent <percent>
#	Files are considered different (returns -1) when this or greater
#	percent of values are different.
#
# Returns -1 if difference is too large or error
#          0 otherwise
#
proc FitsOps:Compare {f1 f2 {reportfile ""} args} {

    # Temp hack fix for comparison regression test
    # combined files now named [rootname]_combined.fits
    # regression test files call them combined_frames.fits
    # Will remove this hack once we fix all regression test makefiles
    # for each sample.
    if {$f1=="combined_frames.fits" && ![file exists $f1]} {
        set flst [glob -nocomplain *combined.fits]
        set f1 [lindex $flst 0]
    }

    # Handle "-diffpercent" arg
    # two values are different if they differ by given percent or greater
    set idx [lsearch $args "-diffpercent"]
    if {$idx!=-1} {
	set val [expr $idx+1]
    	set diff [lindex $args $val]
	set args [lreplace $args $idx $val]
    } else {
	set diff 10.0     ; # default
    }

    # Handle "-failpercent" arg
    # failure occurs when given percent of values are different
    set idx [lsearch $args "-failpercent"]
    if {$idx!=-1} {
	set val [expr $idx+1]
    	set fail [lindex $args $val]
	set args [lreplace $args $idx $val]
    } else {
	set fail 10.00     ; # default 
    }

    # so output will send to stderr and can be seen
    switch $reportfile {
	""        {set fid stdout}
	"stdout"  {set fid stdout}
	"stderr"  {set fid stderr}
	default   {set fid [open $reportfile w]}
    }
	
    puts $fid "Comparing $f1 with $f2:"
	
    # compare file HDUs (aka extensions)
    puts -nonewline $fid "Comparing the extension lists \t-> "
    flush $fid
    set res [FitsOps:CompareExtensionLists $f1 $f2]
    if {$res==0} {
	puts $fid "Ok"
    } else {
	puts $fid "Failed code=$res"
	close $fid
	return $res
    }
	
    # compare file FITS header values
    puts -nonewline $fid "Comparing the Main Header \t-> "
    flush $fid
    set res [FitsOps:CompareHeaders $f1 $f2]
    if {$res==1} {
	puts $fid "Only $f1 SCRUNCHed"
    } elseif {$res==2} {
	puts $fid "Only $f2 SCRUNCHed"
    } elseif {$res==0} {
	puts $fid "Ok"
    } else {
	puts $fid "Failed code=$res"
	close $fid
	return $res
    }
	
    # compare fibre tables
    puts -nonewline $fid "Comparing the Fibre Table Data \t-> "
    flush $fid
    set res [FitsOps:CompareTables $f1 $f2 *FIBRES*]
    if {$res==0} {
	puts $fid "Ok"
    } elseif {$res==-1} {
	puts $fid "At least one file has NO fibre table"
    } else {
	puts $fid "Failed code=$res"
	close $fid
	return $res
    }

    # compare primary image data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid Primary $diff $fail]} {
	return -1
    }

    # compare variance data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid VARIANCE $diff $fail]} {
	return -1
    }

    # compare WAVELEN data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid WAVELA $diff $fail]} {
	return -1
    }

    # compare SHIFTS data
#    if { ! [FitsOps:CompareImage $f1 $f2 $fid SHIFTS $diff $fail]} {
#	return -1
#    }

    # compare THPUT data
#    if { ! [FitsOps:CompareImage $f1 $f2 $fid THPUT $diff $fail]} {
#	return -1
#    }

    # compare SKY data
#    if { ! [FitsOps:CompareImage $f1 $f2 $fid SKY $diff $fail]} {
#	return -1
#    }

    # compare TELCOR data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid TELCOR $diff $fail]} {
	return -1
    }

    # compare SIGMAPRF data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid SIGMAPRF $diff $fail]} {
	return -1
    }

    # compare RWSS data
    if { ! [FitsOps:CompareImage $f1 $f2 $fid RWSS $diff $fail]} {
	return -1
    }

    close $fid
    return 0
}

#******************************************************************************
#
# Same behaviour as FitsOps:Compare except that FIBRE tables are not included
# in the comparison tests, which is a necessary exclusion as currently the 
# fits2ndf and ndf2fits trashes certain unused (for now) columns to random
# gibberish.
#
# Returns -1 if there is a difference, or
#          0 otherwise
#
proc FitsOps:FMOSCompare {f1 f2 {reportfile ""}} {

    # so CruiseControl will send to stderr and can be seen
    switch $reportfile {
	""        {set fid stdout}
	"stdout"  {set fid stdout}
	"stderr"  {set fid stderr}
	default   {set fid [open $reportfile w]}
    }
	
    puts $fid "Comparing $f1 with $f2:"
	
    # compare extensions
    puts -nonewline $fid "Comparing the extension lists \t-> "
    set res [FitsOps:CompareExtensionLists $f1 $f2]
    if {$res==0} {
	puts $fid "Ok"
    } else {
	puts $fid "Failed code=$res"
	close $fid
	return $res
    }
	
    # compare headers
    puts -nonewline $fid "Comparing the Main Header \t-> "
    set res [FitsOps:CompareHeaders $f1 $f2]
    if {$res==1||$res==2||$res==0} {
	puts $fid "Ok"
    } else {
	puts $fid "Failed code=$res"
	close $fid
	return $res
    }
	
    # compare image data
    puts -nonewline $fid "Comparing the Image Data \t-> "
    set stat [FitsOps:CompareImage $f1 $f2 $fid Primary $diff $fail]
    puts -nonewline $fid "${stat}% differ by more than ${diff}% "
    if {$stat<10} {
	puts $fid "Ok"
    } else {
	puts $fid "Failed"
	close $fid
	return -1
    }

    close $fid
    return 0
}

#******************************************************************************
#
# Compares the values of "P" and "S" type fibers only
#
# Returns ???
#
proc CompareFiberImage {f1 f2 percent_diff max_percent} {
 
    set headers [FitsOps:TableDump $f1 *FIBRE* A]
    set type_idx [lsearch $headers "TYPE"]
    incr type_idx 1
    set nfibs $A(n_rows)
    set sum_count 0
    set fib_count 0
    for {set fib 1} {$fib<=$nfibs} {incr fib 1} {
        set fib_type [string toupper $A($fib,$type_idx)]
        if {$fib_type!="P" && $fib_type!="S"} {continue}
        incr fib_count 1
        set fib1_data [FitsOps:GetImageRow $f1 $fib]
        set fib2_data [FitsOps:GetImageRow $f2 $fib]

        # Compare the two lists of data
        set rel_stat [expr 0.01*$percent_diff]
        set max_stat [expr 0.01*$max_percent]
        set n [llength $fib1_data]
        set count 0
        for {set idx 0} {$idx<$n} {incr idx 1} {
        
            set datum1 [lindex $fib1_data $idx]
            set datum2 [lindex $fib2_data $idx]

            # First Sort the case of NULLS
            if {$datum1=="NULL" && $datum2=="NULL"} {continue}
            if {$datum1=="NULL" && $datum2!="NULL"} {incr count 1;continue}            
            if {$datum1!="NULL" && $datum2=="NULL"} {incr count 1;continue}            
            
            # Now look at the absolute/relative difference
            if {[expr abs($datum1)]<1} {
                set stat [expr abs($datum1-$datum2)]
            } else {
                set stat [expr abs(($datum1-$datum2)/$datum1)]
            }
            if {$stat>=$rel_stat} {incr count 1}

        }; # End for idx
        
        set variation_percent  [expr 100.0*$count/$n]
        if {$variation_percent>=$max_percent} {
            puts "Fiber $fib Type=$fib_type \
                  Count=$count ---> ${variation_percent}%"
         }

        
        incr sum_count $count

    };# End for fib
    
    puts "Total variation: [expr 100.0*$sum_count/($n*$fib_count)]"

    
};# End CompareFiberImage

#******************************************************************************
#
# Compares ???
#
# Returns ???
#
proc FitsOps:CompareBadVal {f1 f2} {

   set fname compbadval.txt
   if {[file exists $fname]} {
       file delete $fname
   }
   set fid [open $fname w]
   puts $fid "Comparing Bad Values Between $f1 and $f2"
   set diff_per [FitsOps:CompareImage $f1 $f2 $fid Primary 0.0001 10.0]
   puts $fid "$diff_per differ by more than 0.0001 %"
   close $fid
   if {$diff_per>0.0001} {
      return -1
   }
   return 0
}

#******************************************************************************
#
# Compares the number and names of extensions in the two given FITS files.
#
# Return -1 if number of extensions differ, or
#        -2 if one or more extension names differ, or
#         0 otherwise (all tests match)
#
proc FitsOps:CompareExtensionLists {file1 file2} {

    # Get list of extension names from both files
    set ext_list1 [FitsOps:ListExtensions $file1]
    set ext_list2 [FitsOps:ListExtensions $file2]
    
    # translate names of form "TMP31.MORE.FIBRES" to "FIBRES" in each list
    # former form was created by Starlink 'ndf2fits' progrom 
    # remove "REDUCED" and PSFMAP extensions since they are deprecated
    # ??? Why is SHIFTS extension being ignored ???
    # ignore REDUCTION_ARGS
    set check_list1 ""
    foreach extname $ext_list1 {
	set short_name [lindex [split $extname "."] end]
	if {$short_name!="REDUCED"
	    &&$short_name!="PSFMAP"
	    &&$short_name!="SHIFTS"
	    &&$short_name!="REDUCTION_ARGS"} {
	    lappend check_list1 $short_name
	}
    }
    set check_list2 ""
    foreach extname $ext_list2 {
	set short_name [lindex [split $extname "."] end]
	if {$short_name!="REDUCED"
	    &&$short_name!="PSFMAP"
	    &&$short_name!="SHIFTS"
	    &&$short_name!="REDUCTION_ARGS"} {
	    lappend check_list2 $short_name
	}
    }

    # return -1 if number of extensions is different
    if {[llength $check_list1]!=[llength $check_list2]} {
	return -1
    }	
    
    # return -2 if extension names do not match
    foreach extname $check_list1 {
	if {[lsearch $check_list2 $extname]==-1} {
	    return -2
	}
    }
    
    # here everything checks ok, return 0
    return 0
}

#******************************************************************************
#
# Compare the header data of the two given FITS files.
#
# Return 
#    1 if only $file1 has SCRUNCHED keyword
#    2 if only $file2 has SCRUNCHED keyword
#    0 if ALL TESTS MATCH
#   -1 if number of keywords differ, or
#   -2 if one or more keyword names differ, or
#   -3 if one or more keyword values differ, or
#
proc FitsOps:CompareHeaders {file1 file2} {

    # don't compare values of these keywords
    # all dates and times
    # EXTNAME doesn't appear in some old 2dfdr combined output files
    # WCSNAME appears in some 2df reference files ???
    # CDELT1,CDELT2,CUNIT2 result from NDF -> TDFIO conversion
    # BZERO,BSCALE result from NDF -> TDFIO conversion
    # HDS related elements are added during Starlink conversion
    # LBOUND1,LBOUND2,HDUCLAS1,HDUCLAS2 are added during Starlink conversion
    # PCOUNT,GCOUNT are added by FITSIO
    set skipwords { DATE UTSTART UTEND STSTART STEND
	ZDSTART ZDEND HASTART HAEND
	MEANRA APPRA MEANDEC APPDEC
	UTMJD ELAPSED
	EXTNAME
	WCSNAME
        CDELT* CDELT* CUNIT* CTYPE*
	BZERO BSCALE
        *HDS*
        PCOUNT GCOUNT
	LBOUND1 LBOUND2 HDUCLAS1 HDUCLAS2
	FILEORIG TDFCTEVT GRPNUM OBSNUM RUN
        EQUINOX MJD-OBS DATE-OBS CD1_1 CD2_2
        EPOCH RADECSYS TDFDRVER TDFIOVER 2DFDRVER}

    # read array of values indexed by keyword from each file
    array set header1 [FitsOps:HeaderDump $file1]
    array set header2 [FitsOps:HeaderDump $file2]

    # get keywords from each file into list
    set keywords1 [array names header1]
    set keywords2 [array names header2]

    # remove words not being compared from both lists
    # N.B. array unset does not work under Tcl8.2 used by Starlink
    foreach word $skipwords {
#    array unset header1 *HDS*
#    array unset header2 *HDS*
	while {[set idx [lsearch $keywords1 $word]]!=-1} {
	    set keywords1 [lreplace $keywords1 $idx $idx]
	}
	while {[set idx [lsearch $keywords2 $word]]!=-1} {
	    set keywords2 [lreplace $keywords2 $idx $idx]
	}
    }

    # check for comparing 1st pass with 2nd pass of fibre flat frame
    set idx1 [lsearch $keywords1 SCRUNCH]
    set idx2 [lsearch $keywords2 SCRUNCH]
    set scrunch_flag 0
    if {$idx1!=-1 && $idx2==-1} {
	set scrunch_flag 1
	set keywords1 [lreplace $keywords1 $idx1 $idx1]
    } elseif {$idx1==-1 && $idx2!=-1} {
	set scrunch_flag 2
	set keywords2 [lreplace $keywords2 $idx2 $idx2]
    }

    # return -1 if array sizes differ
    ### puts does not belong at this level ###
    if {[llength $keywords1]!=[llength $keywords2]} {
	puts "$file1 len=[llength $keywords1]"
	puts "$file2 len=[llength $keywords2]"
	return -1
    }
    
    # return -2 if if keyword names differ
    ### puts does not belong at this level ###
    foreach keyword $keywords1 {
	if {[lsearch $keywords2 $keyword]==-1} {
	    puts "keyword=$keyword"
	    return -2
	}
    }

    # here everything checks ok, return 0
    if {$scrunch_flag==0} {
	return 0
    } else {
	return $scrunch_flag
    }
}

#******************************************************************************
#
# Compare the specified table of the two given FITS files.
#
# Return 0 if same, or
#       -1 if one or both files have NO table, or
#       -2 if number of rows differ, or
#       -3 if number of columns differ, or
#       -4 if one or more column names differ, or
#       -5 if one or more values differ
#
proc FitsOps:CompareTables {file1 file2 tablename} {

    # SWITCHFIELD SWITCHPARTNER WLEN columns are added by ndf2fits,fits2ndf
    #   (SWITCHFIELD SWITCHPARTNER are in raw 6df but not AAOmega)
    # SWITCHPARTNER column is being garbaged by ndf2fits,fits2ndf
    set skipcols { SWITCHFIELD SWITCHPARTNER WLEN}

    # check file1 has table
    set fob [fits open $file1 0]     ; # read only
    set hdu [FitsFobOps:FindExtensionHDU $fob $tablename]
    $fob close
    if {$hdu==-1} {
	return -1
    }

    # check file2 has table
    set fob [fits open $file2 0]     ; # read only
    set hdu [FitsFobOps:FindExtensionHDU $fob $tablename]
    $fob close
    if {$hdu==-1} {
	return -1
    }

    # obtain table column names and contents from each file
    set colnames1 [FitsOps:TableDump $file1 $tablename table1A]
    set colnames2 [FitsOps:TableDump $file2 $tablename table2A]
	
    # check number of rows (fibre count) differ
    if {$table1A(n_rows)!=$table2A(n_rows)} {
	return -2
    }

    # remove column names not being compared from both
    foreach name $skipcols {
	while {[set idx [lsearch $colnames1 $name]]!=-1} {
	    set colnames1 [lreplace $colnames1 $idx $idx]
	}
	while {[set idx [lsearch $colnames2 $name]]!=-1} {
	    set colnames2 [lreplace $colnames2 $idx $idx]
	}
    }

    # check number of columns differ
    if {[llength $colnames1]!=[llength $colnames2]} {
	return -3
    }

    # check column names differ
    if {$colnames1!=$colnames2} {
	return -4
    }
    
    # check if values differ
    foreach name $colnames1 {
	set col1 [expr [lsearch $table1A(col_names) $name]+1]
	set col2 [expr [lsearch $table2A(col_names) $name]+1]
	for {set i 1} {$i<=$table1A(n_rows)} {incr i 1} {
	    if {$table1A($i,$col1)!=$table2A($i,$col2)} {
		return -5
	    }
	}
    }
	
    return 0
}

#******************************************************************************
#
# Compare the given arrays of the two given FITS files.
#
# Computes the percent of data for which the relative difference between
# corresponding values exceeds the given (diff) percentage.
#
# file1   - name of first file to be compared
# file1   - name of second file to be compared
# fid     - where output dialog will be sent
# extname - name of HDU to be compared
# diff    - percent two values must differ by to be called different
# fail    - failure is indicated if more than this percent are different 
#
# Returns  1 ok, percent is less than or equal to the given fail percentage
#          0 failed, percent is greater than the given fail percentage
#         -1 if both files do not have given $extname
#         -2 if $file1 does not have given $extname
#         -3 if $file2 does not have given $extname
#         -4 if array sizes differ
#
# Note this is done using Tcl lists and foreach which are slow. We are not
# using the vexpr command whch would be quicker as there is currently a
# problem with using this command within drcontrol.
#
proc FitsOps:CompareImage {file1 file2 fid extname diff fail} {

    # open files
    set fob1 [fits open $file1 0]  ;# read only
    set fob2 [fits open $file2 0]  ;# read only

    # get index number for given HDU
    # add universal match (*) as prefit to handle NDF names
    if {$extname=="Primary"} { 
	set hdu1 1
	set hdu2 1
    } else {
	set searchname "*${extname}"
	set hdu1 [FitsFobOps:FindExtensionHDU $fob1 $searchname]
	set hdu2 [FitsFobOps:FindExtensionHDU $fob2 $searchname]
    }

    # check HDUs exist
    if {$hdu1==-1&&$hdu2==-1} {
	puts $fid "NO $extname HDU in either file"
	return 1
    } elseif {$hdu1==-1} {
	puts $fid "Only $f2 has $extname"
	return -2
    } elseif {$hdu2==-1} {
	puts $fid "Only $f1 has $extname"
	return -3
    }

    puts -nonewline $fid "Comparing the $extname Data \t-> "
    flush $fid

    # move to given HDU
    $fob1 move $hdu1
    $fob2 move $hdu2

    # check image sizes
    if {[$fob1 info imgdim]!=[$fob2 info imgdim]} {
	puts $fid "File array sizes differ"
	return -4
    }

    # load image data into tcl lists
    set image_list1 [$fob1 get image]
    $fob1 close
    set image_list2 [$fob2 get image]
    $fob2 close
    
    # compare image data
    set rel_stat [expr 0.01*$diff]
    set count 0
    set n [llength $image_list1]

    for {set idx 0} {$idx<$n} {incr idx 1} {
	set datum1 [lindex $image_list1 $idx]
	set datum2 [lindex $image_list2 $idx]
	if {$datum1=="NULL" || $datum2=="NULL"} {
	    if {$datum1!=$datum2} {
		incr count 1
	    }
	    continue
	}
	if {[expr abs($datum1)]<1} {
	    set stat [expr abs($datum1-$datum2)]
	} else {
	    set stat [expr abs(($datum1-$datum2)/$datum1)]
	}
	if {$stat>=$rel_stat} {
	    incr count 1
	}
    }	
	
    set percent [expr 100.0*$count/$n]
    puts -nonewline $fid "${percent}% differ by more than ${diff}% "
    if {$percent<=$fail} {
	puts $fid "Ok"
	return 1
    } else {
	puts $fid "Failed"
	return 0
    }
}

#******************************************************************************
#
# Returns FITS header data as a Tcl array for the given FITS file. Blank and 
# HISTORY records are excluded.
#
# ??? Why not use "objName get keyword" ???
#
# Arguments
#   fitsfile  - FITS filename
#
# Optional Arguments (args)
#  -c
#	Populate dataArray with comments instead of values
#  -extension [extension name] 
#	Make Header dump for specified extension
#	Extension name may include wildcards, eg *.FIBRES
#
# Returns Tcl (associative) array where keyword values (or comments) are
#     indexed by keyword names, i.e. value(keyword name)
#
proc FitsOps:HeaderDump {fitsfilename args} {

    # parse include comments argument (-c)
    set idx [lsearch $args "-c"]
    if {$idx!=-1} {
	set comment_flag 1
	set args [lreplace $args $idx $idx]
    } else {
	set comment_flag 0
    }

    # parse extension name argument (-extension)
    set idx [lsearch $args "-extension"]
    if {$idx!=-1} {
	set val [expr $idx+1]
	set extname [lindex $args $val]
	set args [lreplace $args $idx $val]
    } else {
	set extname ""
    }

    # open FITS file for read only
    set fits_obj [fits open $fitsfilename 0]

    # move to appropriate extension (if given)
    if {$extname!=""} {
        # find HDU with extension name that matches extensioname
        set hdu [FitsFobOps:FindExtensionHDU $fits_obj $extname]
        if {$hdu==-1} {
	    $fits_obj close; return ""
	}
	$fits_obj move $hdu
    }

    # go through all keyword records in extension and dump to array
    set keyname_list ""
    for {set idx 1} {$idx<[$fits_obj info nkwds]} {incr idx 1} {
    
	set record [$fits_obj get keyword -num $idx]
		
	set keyname [lindex $record 0]
	set value   [lindex $record 1]
	set comment [lindex $record 2]

	# don't include blank records
	if {[string is space $keyname]==1} {
	    continue
	}

	# don't include history records
	if {[string toupper $keyname]=="HISTORY"} {
	    continue
	}

	# set array element to either value or comment indexed by keyname
	if {$comment_flag} {
 	    set dataA($keyname) $comment
	} else {
	    set dataA($keyname) [string trim $value " '"]
	}
    }

    # close FITS file
    $fits_obj close
    
    ### puts does not belong at this level ###
    foreach name [array names dataA] {
	if {[string is space $name]==1} {
	    puts "dump bang"
	}
    }

    # return array of values indexed by names
    return [array get dataA]
}



proc FitsOps:CompareVariance {file1 file2 outfile} {

    # ---------
    # COMMENTS:
    # ---------
    # Routine for comparing the variance values of two extracted
    # frames. This is to determine the validity of variance
    # propgation formulation with some empirical/stochastic
    # determined values. We measure various statistics for
    # each cell, i.e., diff, abs diff, rel diff and ratio.
    # and place in a frequency table which is then outputed
    # to outfile


    # -------------------------------------------------
    # Open fits files and locate the VARIANCE extension
    # -------------------------------------------------
    set fob1 [fits open $file1 0]  ;# read only
    set fob2 [fits open $file2 0]  ;# read only

    set nbins 200
    set binwidth 1.0 ;# 1% binwidths

    # Get index number for given HDU
    # add universal match (*) as prefit to handle NDF names
    set extname "VARIANCE"
    if {$extname=="Primary"} { 
	set hdu1 1
	set hdu2 1
    } else {
	set searchname "*${extname}"
	set hdu1 [FitsFobOps:FindExtensionHDU $fob1 $searchname]
	set hdu2 [FitsFobOps:FindExtensionHDU $fob2 $searchname]
    }


    # --------------
    # Sanity Checks:
    # --------------
    # check HDUs exist
    if {$hdu1==-1&&$hdu2==-1} {
	puts  "NO $extname HDU in either file"
	return 1
    } elseif {$hdu1==-1} {
	puts  "Only $f2 has $extname"
	return -2
    } elseif {$hdu2==-1} {
	puts  "Only $f1 has $extname"
	return -3
    }

    puts -nonewline "Comparing the $extname Data \t-> "
   

    # move to given HDU
    $fob1 move $hdu1
    $fob2 move $hdu2

    # check image sizes
    if {[$fob1 info imgdim]!=[$fob2 info imgdim]} {
	puts  "File array sizes differ"
	return -4
    }
    

    # -------------------------------------
    # Get the variance data from both files
    # -------------------------------------
    # load image data into tcl lists
    set image_list1 [$fob1 get image]
    $fob1 close
    set image_list2 [$fob2 get image]
    $fob2 close
    
    
    # -----------------------
    # Build Frequency tables:
    # -----------------------
    set Bcount 0
    set count 0
    set n [llength $image_list1]
    for {set f_idx 1} {$f_idx<=$nbins} {incr f_idx 1} {set TableV($f_idx) 0}

    set sum1 0.0
    set sum2 0.0
    for {set idx 0} {$idx<$n} {incr idx 1} {
	set datum1 [lindex $image_list1 $idx]
	set datum2 [lindex $image_list2 $idx]
	if {$datum1=="NULL" || $datum2=="NULL"} {
	    if {$datum1!=$datum2} {
		incr Bcount 1
	    }
	    continue
	}
	set diff     [expr ($datum1-$datum2)]
        set abs_diff [expr abs($diff)]
        set rel_val  [expr $datum1/$datum2]
	set rel_err  [expr $abs_diff/$datum2]
        
        #set opt rel_err
        set opt rel_val
        switch $opt {
            "rel_err" {set f_idx [expr int ($rel_err*100.0/$binwidth) + 1]}
            "rel_val" {set f_idx [expr int ($rel_val*100.0/$binwidth) + 1]}
        }
        incr TableV($f_idx) 1
        incr count 1
        set sum1 [expr $sum1+$rel_err]
        set sum2 [expr $sum2+$rel_val]
        
    }
    
    puts "Total=$n, Bcount=$Bcount count=$count"
    puts "Average Relative Error=[expr $sum1/$count]"
    puts "Average Relative Value=[expr $sum2/$count]"
    set fid [open $outfile w]
    for {set f_idx 1} {$f_idx<=$nbins} {incr f_idx 1} {
        puts $fid "[expr ($f_idx-1)*$binwidth] 0.0"
        puts $fid "[expr ($f_idx-1)*$binwidth] $TableV($f_idx)"
        puts $fid "[expr ($f_idx)  *$binwidth] $TableV($f_idx)"
        puts $fid "[expr ($f_idx)  *$binwidth]  0.0"
    }
    close $fid
	
};# FitsOps:CompareVariance

