################################################################################
#
# AAOmega-IFU Reconstruction Map Tool
#
# See HelpPanel below for a description of this tool.
#
# Future Development Thoughts
# 1. Scale options-, absolute min and max, percentile or median+-n*sigma
# MNB Mon Jan 8th 2007
#
################################################################################

# Define Global variables
global IfuMap

# AAOmega-IFU has 16 groups of 32 closely packed fibres
set IfuMap(nrows) 32        ;# rows in AAOmega-IFU lenset grid
set IfuMap(ncols) 16        ;# columns in same

# dimensions in screen units of each grid tile
set IfuMap(tile_width) 20  ;# 25
set IfuMap(tile_height) 20  ;# 25

set IfuMap(colour,grid) slategrey   ;# grid lines colour
set IfuMap(colour,text) cyan
set IfuMap(colour,highlight) yellow ;# black ;#slategrey

set IfuMap(root) .ifuMap

set IfuMap(pnames,1) "P1"
set IfuMap(pnames,2) "P2"
set IfuMap(pnames,3) "P3"
set IfuMap(pnames,4) "P4"

# Setup Tk and load up tclGraph
puts "Loading tclGraph.tcl package"
package require Tk
source [file join $env(DRCONTROL_DIR) tclGraph.tcl]

# If there are no widgets in place then hide the main window ???
if {[winfo children .]==""} {
    wm withdraw .
}

#******************************************************************************
#
# Main (execution begins here)
#
# Arguments
#   red_fitsfile    - input reduced IFU file
#
proc IfuMap { { redfile "" } } {

    global IfuMap
    global IfuMap:dataA
    global IfuMap:sdataA

    # if reduced filename was not passed, put up panel to obtain it from user
    if {$redfile==""} {
        set file_types {{"Reduced file" {"*red.fits"}}}
        set redfile [tk_getOpenFile -filetypes $file_types \
                        -title "Reduced IFU File to Map"]
    }

    # quit if user chose cancel
    if {$redfile==""} {
        return
    }

    # create our main panel
    IfuMap:CreateWindow

    # read input file
    set IfuMap(reduced_file) $redfile
    IfuMap:ReadRedFile $redfile IfuMap:dataA

    # get scaled data for display
    IfuMap:GenScaleData 0.0 65536.0 IfuMap:dataA IfuMap:sdataA

    # configure tiles
    IfuMap:ConfigTiles IfuMap:sdataA IfuMap:dataA
}

#******************************************************************************
#
# Creates the main IfuMap panel
#
# This panel contains the frames Canvas, File, Display, Plot, Edit and Status.
#
proc IfuMap:CreateWindow { } {

    global IfuMap
    global IfuMapSelection

    set root $IfuMap(root)
    destroy $root
    toplevel $root

    # CANVAS frame to hold 32 x 16 grid tiles
    set nrows $IfuMap(nrows)
    set ncols $IfuMap(ncols)
    set tile_width $IfuMap(tile_width)
    set tile_height $IfuMap(tile_height)
    set width [expr $tile_width*$ncols]
    set height [expr $tile_height*$nrows] 

    set c $root.canvas
    canvas $c -width $width -height $height -bg grey
    grid $c -row 0 -column 0

    # process next grid tile row each time around this loop
    set ylast [expr ($nrows+1)*$tile_height]
    for {set i 1} {$i<=$nrows} {incr i 1} {
        set y0 [expr ($nrows-$i)*$tile_height]
        set y1 [expr $y0+$tile_height]
        if {$i!=1} {
            set x0 0
            set x1 [expr $ncols*$tile_width]
            $c create line $x0 $y1 $x1 $y1 \
                -fill $IfuMap(colour,grid) -stipple gray50 -tags grid
        }

        # process next grid tile column each time around this loop
        for {set j 1} {$j<=$ncols} {incr j 1} {
            set x0 [expr ($j-1)*$tile_width]
            set x1 [expr $j*$tile_width]
            if {$j!=$ncols} {
                $c create line $x1 0 $x1 $ylast \
                    -fill $IfuMap(colour,grid) -stipple gray50 -tags grid
            }

            set tile ${i}:${j}
            set tile_text ${tile}:type
            $c create rectangle $x0 $y0 $x1 $y1 -tags "$tile rectangle"
            set xm [expr 0.5*($x0+$x1)]
            set ym [expr 0.5*($y0+$y1)]
            $c create text $xm $ym -anchor c -text "" \
                -fill $IfuMap(colour,text) -tags "$tile_text text"

            # setup tile action binds
            $c bind $tile <ButtonPress-1> "IfuMap:TileBindB1 $tile"
            $c bind $tile_text <ButtonPress-1> "IfuMap:TileBindB1 $tile"
            $c bind $tile <Double-ButtonPress-1> \
                "IfuMap:TileBindDoubleB1 $tile"
            $c bind $tile_text <Double-ButtonPress-1> \
                "IfuMap:TileBindDoubleB1 $c $tile"
            $c bind $tile <Control-ButtonPress-1> \
                "IfuMap:TileBindShiftB1 $tile"
            $c bind $tile_text <Control-ButtonPress-1> \
                "IfuMap:TileBindShiftB1 $tile"
            $c bind $tile <Any-Enter> "IfuMap:TileBindEnter $tile"
            $c bind $tile_text <Any-Enter> "IfuMap:TileBindEnter $tile"
            $c bind $tile <Any-Leave> "IfuMap:TileBindLeave"

            # initial tile type
            IfuMap:SetTileType $tile ""
        }
    }

    # no tiles selected at this point
    set IfuMapSelection ""

    # bind canvas mouse actions
    bind $c <B1-Motion> "IfuMap:CanvasBindB1Motion %x %y"
    bind $c <ButtonRelease-1> { set IfuMap(last_tile) "" }
    $c bind grid <ButtonPress-1> { IfuMap:SetSelection "" }
    
    # bframe contains Display, Plot, Edit and File frames
    set bframe $root.bframe
    frame $bframe -relief raised -bd 1
    grid $bframe -row 0 -column 1 -sticky ns

    # Display frame contains Grid and Text checkboxs, and GreyScale button
    label $bframe.l0 -text "Display"

    # checkbox to toggle display of grid lines
    checkbutton $bframe.grid -text Grid -relief raised \
        -variable IfuMap(disvar,grid) -command "IfuMap:ToggleGrid $c grid"
    set IfuMap(disvar,grid) 1  ;# initially on

    # checkbox to toggle display of grid box text
    checkbutton $bframe.text -text Text -relief raised \
        -variable IfuMap(disvar,text) -command "IfuMap:ToggleText $c text"
    set IfuMap(disvar,text) 1  ;# initially on

    # button to put up GreyScale statistics panel
    button $bframe.greyscale -text Greyscale... -command IfuMap:GreyScalePanel

    # pack Display frame members into $bframe
    pack $bframe.l0 $bframe.grid  $bframe.text $bframe.greyscale \
        -side top -fill x

    frame $bframe.j1 -height 30    ;# some space

    # PLOT frame contains plotting buttons
    label $bframe.lp -text "Plot"
    button $bframe.plot1 -text "Plot P1" \
        -command "IfuMap:PlotProg P1" -state disabled
    button $bframe.plot2 -text "Plot P2" \
        -command "IfuMap:PlotProg P2" -state disabled
    button $bframe.plot3 -text "Plot P3" \
        -command "IfuMap:PlotProg P3" -state disabled
    button $bframe.plot4 -text "Plot P4" \
        -command "IfuMap:PlotProg P4" -state disabled
    button $bframe.plotxycut -text "Plot XY Cut" \
        -command "IfuMap:PlotXYCut" -state disabled

    # pack Plot frame members into $bframe
    pack $bframe.lp $bframe.plot1 $bframe.plot2 $bframe.plot3 $bframe.plot4 \
        $bframe.plotxycut -side top -fill x
        
    # EDIT frame contains Edit, Plot1-4, Sky, Undo and Unused buttons
    frame $bframe.j2 -height 30    ;# some space
    label $bframe.le -text "Edit"
    button $bframe.prog1 -text "Prog1" \
        -command "IfuMap:SetSelectionToType P1" -state disabled
    button $bframe.prog2 -text "Prog2" \
        -command "IfuMap:SetSelectionToType P2" -state disabled
    button $bframe.prog3 -text "Prog3" \
        -command "IfuMap:SetSelectionToType P3" -state disabled
    button $bframe.prog4 -text "Prog4" \
        -command "IfuMap:SetSelectionToType P4" -state disabled
    button $bframe.sky -text "Sky " \
        -command "IfuMap:SetSelectionToType S" -state disabled
    button $bframe.undef -text "Undef" \
        -command "IfuMap:SetSelectionToType \"\"" -state disabled
    button $bframe.unused -text "Unused" \
        -command "IfuMap:SetSelectionToType U" -state disabled

    # pack Edit frame members into $bframe
    pack $bframe.le $bframe.prog1 $bframe.prog2 $bframe.prog3 $bframe.prog4 \
        $bframe.sky $bframe.unused $bframe.undef -side top -fill x

    # File frame contains buttons to Read and Save ASCII text file
    label $bframe.lf -text "File"
    button $bframe.read -text "Read" -command {IfuMap:ReadAsciiFile ""; \
        IfuMap:ConfigTiles IfuMap:sdataA IfuMap:dataA }
    button $bframe.save -text "Save" \
        -command "IfuMap:WriteAsciiFile"

    # pack File frame members into $bframe
    pack $bframe.lf $bframe.read $bframe.save -side top -fill x

    # Help button and pack into $bframe
#    button $bframe.help -text "Help" -command {IfuMap:HelpPanel}
#    pack $bframe.help -side bottom -fill x
        
    # set IfuMap variables
    set IfuMap(path,prog_button1) $bframe.prog1
    set IfuMap(path,prog_button2) $bframe.prog2
    set IfuMap(path,prog_button3) $bframe.prog3
    set IfuMap(path,prog_button4) $bframe.prog4
    set IfuMap(path,sky_button) $bframe.sky
    set IfuMap(path,undef_button) $bframe.undef 
    set IfuMap(path,dead_button) $bframe.unused
    set IfuMap(path,plot,P1) $bframe.plot1
    set IfuMap(path,plot,P2) $bframe.plot2
    set IfuMap(path,plot,P3) $bframe.plot3
    set IfuMap(path,plot,P4) $bframe.plot4
    set IfuMap(path,plotxycut_button) $bframe.plotxycut
    set IfuMap(path,save_button) $bframe.save 
    set IfuMap(plot_buttons) "plot1 plot2 plot3 plot4"
    set IfuMap(canvas) $c
    
    # STATUS frame where tile information is displayed
    label $root.status -textvariable IfuMap(disvar,display) -bg white
    pack $root.status -side left -fill x
    grid $root.status -row 1 -column 0 -columnspan 2 -sticky we
       
    # last tile where mouse moved
    set IfuMap(last_tile) ""
}

#******************************************************************************
#
# Handler for the "Grid" checkbutton.
# Toggles grid line display on given canvas
#
# Arguments
#   canvas   -
#   grid_tab - 
#
proc IfuMap:ToggleGrid {canvas grid_tag} {

    global IfuMap

    # set grid line colour to either defined colour or invisible
    if {$IfuMap(disvar,grid)} {
        set colour $IfuMap(colour,grid)
    } else {
        set colour ""
    }

    # write grid lines
    $canvas itemconfig $grid_tag -fill $colour
}

#******************************************************************************
#
# Handler for the "Text" checkbutton.
# Toggles grid box text display on given canvas
#
# Arguments
#   canvas   -
#   grid_tab - 
#
proc IfuMap:ToggleText {canvas text_tag} {

    global IfuMap

    # set text colour to either defined colour or invisible
    if {$IfuMap(disvar,text)} {
        set colour $IfuMap(colour,text)
    } else {
        set colour ""
    }

    # write text
    $canvas itemconfig $text_tag -fill $colour
}

#******************************************************************************
#
# Handler for the "Plot P1", etc. plotting buttons
# Plots combined spectra of selected Program type
#
# Arguments
#   match_type - "P1", "P2", "P3" or "P4"
#
proc IfuMap:PlotProg {match_type} {     

    global IfuMap
    
    # create $rowlist of fibre numbers with matching type
    set rowlist ""
    for {set i 1} {$i<=$IfuMap(nrows)} {incr i 1} {
        for {set j 1} {$j<=$IfuMap(ncols)} {incr j 1} {
            set tile ${i}:${j}
            if {[IfuMap:GetTileType $tile]!=$match_type} {
                continue   ;# ignore non-matching type
            }
            puts "tile=$tile"
            lappend rowlist [IfuMap:Tile2FibreNo $tile]
        }
    }
    puts $rowlist
        
    # do plot
    IfuMap:ViewSumRows $IfuMap(reduced_file) $rowlist
}

#******************************************************************************
#
# Handler for the "Greyscale..." button.
# Creates the GreyScale statistics panel
#
proc IfuMap:GreyScalePanel { } {

    global IfuMap
    global IfuMapPanel
    global IfuMap:sdataA
    
    set mainroot $IfuMap(root)
    set root $mainroot.greyscalePanel
    set IfuMapPanel(root) $root
    destroy $root
    toplevel $root
    wm transient $root $mainroot
    wm title $root "Grey Scale Settings"
        
    set statframe $root.statframe
    set clipframe $root.clipframe
    set buttonframe $root.buttonframe
    frame $statframe -relief raised -bd 2
    frame $clipframe -relief raised -bd 2
    frame $buttonframe -relief raised -bd 2
    pack $statframe $clipframe $buttonframe -fill x \
        -expand 1 -ipadx 2m -ipady 2m
        
    # Statistic info display frame
    label $statframe.l -text "Statistics"
    text  $statframe.text -bg white -yscrollcommand "$statframe.sb set"
    scrollbar $statframe.sb -command  "$statframe.text yview"
    grid  $statframe.l -row 0 -column 0 -sticky w
    grid $statframe.text -row 1 -column 0
    grid $statframe.sb -row 1 -column 1 -sticky ns
    set text $statframe.text
        
    # frame for Clipped Min and Clipped Max entry, and Stat Clipped... button
    label $clipframe.lmin -text "Clipped Min"
    entry $clipframe.emin -textvariable IfuMapPanel(disvar,clipmin)
    label $clipframe.lmax -text "Clipped Max"
    entry $clipframe.emax -textvariable IfuMapPanel(disvar,clipmax)
    button $clipframe.stat -text "Stat Clipped..." \
        -command IfuMap:GreyScalePanelStatClippedData
    pack $clipframe.lmin $clipframe.emin $clipframe.lmax $clipframe.emax \
        -side left
    pack $clipframe.stat -side left -padx 4m 
        
    # frame for apply and close buttons
    button $buttonframe.apply -text Apply -width 7 \
        -command IfuMap:GreyScalePanelApply
    button $buttonframe.close -text Close -width 7 -command "destroy $root"
    pack  $buttonframe.apply $buttonframe.close -side left -fill x -padx 4m

    set IfuMapPanel(disvar,clipmin) [set IfuMap:sdataA(used_min)]
    set IfuMapPanel(disvar,clipmax) [set IfuMap:sdataA(used_max)]

    $text insert end "Full data range statistics:\n"
    IfuMap:GreyScalePanelDisplayStats $text

    set IfuMapPanel(path,text) $text
}

#******************************************************************************
#
# Handler for the GreyScale panel "Stat Clipped..." button.
# Writes clipped statistics in GreyScale statistics panel
#
proc IfuMap:GreyScalePanelStatClippedData { } {

    global IfuMapPanel
                
    set clipmin $IfuMapPanel(disvar,clipmin)
    set clipmax $IfuMapPanel(disvar,clipmax)
    set textpath $IfuMapPanel(path,text)
        
    $textpath insert end \n
    $textpath insert end \
        "Clipped data statistics From: $clipmin To: $clipmax\n"
    IfuMap:GreyScalePanelDisplayStats $textpath 1 $clipmin $clipmax
    $textpath yview end
}

#******************************************************************************
#
# Handler for the GreyScale panel "Apply" button.
# Does ???
#
proc IfuMap:GreyScalePanelApply { } {

    global IfuMap
    global IfuMapPanel
    global IfuMap:dataA
    global IfuMap:sdataA

    # get min from panel and check
    set clipmin $IfuMapPanel(disvar,clipmin)
    if {![string is double $clipmin]} {
        set IfuMapPanel(disvar,clipmin) ""
        return
    }
        
    # get max from panel and check
    set clipmax $IfuMapPanel(disvar,clipmax)
    if {![string is double $clipmax]} {
        set IfuMapPanel(disvar,clipmax) ""
        return
    }

    # check min < max
    if {$clipmax<$clipmin} {
        set IfuMapPanel(disvar,clipmin) ""
        set IfuMapPanel(disvar,clipmax) ""
        return
    }

    # create scaled version of data using min and max values
    IfuMap:GenScaleData $clipmin $clipmax IfuMap:dataA IfuMap:sdataA

    # ???
    IfuMap:ColourTiles IfuMap:sdataA
} 

#******************************************************************************
#
# Displays data statistics in GreyScale panel
#
# Arguments
#   textpath  - ???
#   clip_flag - optional whether to apply clipmin & clipmax
#   clipmin   - optional remove data below this value
#   clipmax   - optional remove data above this value
#
proc IfuMap:GreyScalePanelDisplayStats { \
  textpath { clip_flag 0 } { clipmin "" } { clipmax "" } } {

    global IfuMap:dataA
        
    # get data & clip if requested
    set lst [set IfuMap:dataA(list)]
    if {$clip_flag} {
        set lst1 [set IfuMap:dataA(sorted_list)]
        set idx1 [vexpr SUM(lst1<$clipmin)]
        set idx2 [vexpr SUM(lst1<=$clipmax)]
        set lst2 [lrange $lst1 $idx1 $idx2]
    } else {
        set lst2 [set IfuMap:dataA(sorted_list)]
    }
                
    set min [lindex $lst2 0]
    set max [lindex $lst2 end]
    set n [llength $lst2]
    set data_pc [expr 100*$n/[llength $lst]]
        
    $textpath insert end "n=$n (${data_pc}%) \n"
    $textpath insert end "Range: \[ $min , $max \]\n"
    
    set avr [vexpr AVERAGE(lst2)]
    set std [vexpr STDDEV(lst2)]
    set med [vexpr MEDIAN(lst2)]
    $textpath insert end "AVR=$avr STD=$std MED=$med\n"
    
    $textpath insert end \n
    
    set lst3 $lst2
    set n [llength $lst3]
    set pc_list {1 10 95 , 2 30 96 , 3 50 97 , 4 70 98 , 5 90 99} 
    foreach pc $pc_list {
        if {$pc==","} {
            $textpath insert end \n
            continue
        }

        set idx [expr int(0.01*$pc*$n)]
        set val [lindex $lst3 $idx]
        set val [format "%f" $val]
        $textpath insert end "$pc%=$val\t"
    }
    $textpath insert end \n\n
        
    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # statistics below are NOT currently shown
    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    return
        
    set pidx [vexpr SUM(lst2<=0.0)]
    set lst3 [lrange $lst2 $pidx end]

    puts "[lindex $lst3 0] - [lindex $lst3 end]"
    puts "n  =[llength $lst3]"
    puts "AVR=[vexpr AVERAGE(lst3)]"
    puts "STD=[vexpr STDDEV(lst3)]"
    puts "MED=[vexpr MEDIAN(lst3)]"
        
    set pc_list {0 10 20 30 40 50 60 70 80 90 95 96 97 98 99}
    set n [llength $lst3]
    set file tmp_dist.dat
    if {[file exists $file]} {
        file delete $file
    }
    set fid [open $file w]
    foreach pc $pc_list {
        set idx [expr int(0.01*$pc*$n)]
        set val [lindex $lst3 $idx]
        puts "$pc%=$val"
        puts $fid "$val $pc"
    }
    close $fid
        
    set med [vexpr MEDIAN(lst3)]
    set l3  [vexpr abs(lst3-$med)]
    set meddev [vexpr MEDIAN(l3)]
    puts "$med $meddev"
    
    set n [llength $lst3]
    set idx [expr int($n/2)]
    set lst4 [lrange $lst3 0 $idx]
    puts "n  =[llength $lst4]"
    puts "AVR=[vexpr AVERAGE(lst4)]"
    puts "STD=[vexpr STDDEV(lst4)]"
    puts "MED=[vexpr MEDIAN(lst4)]"
    
    set lst4 [lrange $lst3 $idx end]
    puts "n  =[llength $lst4]"
    puts "AVR=[vexpr AVERAGE(lst4)]"
    puts "STD=[vexpr STDDEV(lst4)]"
    puts "MED=[vexpr MEDIAN(lst4)]"
        
#    set l3 [vexpr abs(lst2)]
#    set meddev [vexpr MEDIAN(l3)]
#    puts "meddev=$meddev"
        
    set idx [expr int ([llength $lst2]*0.95)]
    set max [lindex $lst2 $idx]
    #set min [lindex $lst2 0]
    #set max [lindex $lst2 end]
    set idx [expr int ([llength $lst2]*0.98)]
    set use_max [lindex $lst2 $idx]
    set idx [expr int ([llength $lst2]*0.02)]
    set use_min [lindex $lst2 $idx]
    set use_max 26
}

#******************************************************************************
#
# Handler for a Mouse Button Press in grid tile area
# Does ???
#
# Arguments
#   tile   - tile under mouse pointer
#
proc IfuMap:TileBindB1 { tile } {

    # select given tile
    IfuMap:SetSelection $tile
}

#******************************************************************************
#
# Handler for a Mouse Button Double-Click in grid tile area
# Does ???
#
# Arguments
#   tile   - tile under mouse pointer
#
proc IfuMap:TileBindDoubleB1 { tile  } {

    global IfuMap
    
    # get fibre number corresponding to given tile
    set fibreno [IfuMap:Tile2FibreNo $tile]

    set redfile $IfuMap(reduced_file)

    puts "FitsOps:ViewSumRows $redfile $fibreno"
    set result [IfuMap:ViewSumRows $redfile "$fibreno"]
    if {$result!=""} {
        set msg "Unable to plot for fibre $fibreno"
        #tk_message -message $msg -type ok
        puts $msg
    }
}

#******************************************************************************
#
# Handler for a Control-Mouse Button Press in grid tile area
# Does ???
#
# Arguments
#   tile   - tile under mouse pointer
#
proc IfuMap:TileBindShiftB1 { tile } {

    global IfuMapSelection

    # add given tile to current selection set
    set selset $IfuMapSelection
    if {[lsearch $selset $tile]==-1} {
        IfuMap:AppendSelection $tile
    } else {
        IfuMap:SetSelection [lreplace $selset $idx $idx]
    }
}

#******************************************************************************
#
# Handler for Mouse In grid tile area
# Does ???
#
# Arguments
#   tile   - tile under mouse pointer
#
proc IfuMap:TileBindEnter { tile } {

    global IfuMap
    global IfuMapTileValues

    # write status line
    # first split tile index into row and column
    set rowcol [split $tile :]
    set xgpos [lindex $rowcol 1]
    set ygpos [lindex $rowcol 0]

    # get tile type, set to "U" if none
    set type [IfuMap:GetTileType $tile]
    if {$type==""} {
        set type "U"
    }

    set value $IfuMapTileValues($tile)
    set fibno [IfuMap:Tile2FibreNo $tile]
    set IfuMap(disvar,display) [format "XGPOS=$xgpos YGPOS=$ygpos \
        Type=$type Value=%4.2f Fibre=$fibno" $value]
}

#******************************************************************************
#
# Handler for Mouse Leave grid tile area
# Does ???
#
proc IfuMap:TileBindLeave { } {

    global IfuMap

    # blank status line
    set IfuMap(disvar,display) ""
}

#******************************************************************************
#
# Handler for Mouse Movement in canvas frame action
# Does ???
#
# Arguments
#   x      - X position in canvas
#   y      - Y position in canvas
#
proc IfuMap:CanvasBindB1Motion { x y } {

    global IfuMap

    # derive tile under mouse position
    set ix [expr int($x/$IfuMap(tile_width))+1]
    set iy [expr $IfuMap(nrows)-int($y/$IfuMap(tile_height))]
    set tile ${iy}:${ix}

    # ignore if tile has not changed
    if {$IfuMap(last_tile)==$tile} {
        return
    }

    set tile_chain [IfuMap:GetTileChain $IfuMap(last_tile) $tile]
    foreach chain $tile_chain {
        IfuMap:AppendSelection $chain
    }
    set IfuMap(last_tile) $tile
}

#******************************************************************************
#
# ???
#
# Arguments
#   tile1 -
#   tile2 -
#
# Returns ???
#
proc IfuMap:GetTileChain { tile1 tile2 } {
        
    if {$tile1==$tile2 || $tile1==""} {
        return [list $tile2]
    }
        
    set rowcol1 [split $tile1 :]
    set rowcol2 [split $tile2 :]
    set row1 [lindex $rowcol1 0]
    set col1 [lindex $rowcol1 1]
    set row2 [lindex $rowcol2 0]
    set col2 [lindex $rowcol2 1]
        
    set chain ""
    
    if {$col1==$col2} {
        if {$row2<$row1} {
            set tmp $row1
            set row1 $row2
            set row2 $tmp
        }
        for {set row $row1} {$row<=$row2} {incr row 1} {
            set tile ${row}:${col1}
            lappend chain $tile
        }
                
        return $chain
    }
        
    if {$row1==$row2} {
        if {$col2<$col1} {
            set tmp $col1
            set col1 $col2
            set col2 $tmp
        }
        if {$col2>$col1} {
            set step 1
        } {
            set step -1
        }
        for {set col $col1} {$col<=$col2} {incr col $step} {
            set tile ${row1}:${col}
            lappend chain $tile
        }
                
        return $chain
    }
        
    return [list $tile1 $tile2]
}

#******************************************************************************
#
# Sets current selection to given tile or list ???
#
# Arguments
#   tile_list -
#
proc IfuMap:SetSelection { tile_list } {

    global IfuMapSelection

    # remove highlight from each currently selected tile
    foreach tile $IfuMapSelection {
        IfuMap:DeHighlight $tile
    }

    # set new selection list
    set IfuMapSelection $tile_list

    # now highlight each new selected tile
    foreach tile $IfuMapSelection {
        IfuMap:Highlight $tile
    }

    # ???
    if {[llength $IfuMapSelection]==0} {
        IfuMap:DisableEditButtons
    } else {
        IfuMap:EnableEditButtons
    }

    # enable or disable plot XY cut button
    if {[llength $IfuMapSelection]==1} {
        IfuMap:EnablePlotXYCutButton
    } else {
        IfuMap:DisablePlotXYCutButton
    }
}

#******************************************************************************
#
# ???
#
# Arguments
#   tile -
#
proc IfuMap:AppendSelection { tile } {

    global IfuMapSelection
        
    if {[lsearch $IfuMapSelection $tile]==-1} {
        lappend IfuMapSelection $tile
        IfuMap:Highlight $tile
    }
    if {[llength $IfuMapSelection]==0} {
        IfuMap:DisableEditButtons
    } else {
        IfuMap:EnableEditButtons
    }

    # enable or disable plot XY cut button
    if {[llength $IfuMapSelection]==1} {
        IfuMap:EnablePlotXYCutButton
    } else {
        IfuMap:DisablePlotXYCutButton
    }
}

#******************************************************************************
#
# Highlights the given tile
#
# Arguments
#   tile - tile index
#
proc IfuMap:Highlight { tile } {

    global IfuMap

    set canvas $IfuMap(canvas)

    $canvas itemconfig $tile -outline $IfuMap(colour,highlight)
    $canvas raise $tile
    $canvas raise text
}

#******************************************************************************
#
# Removes highlight from the given tile
#
# Arguments
#   tile - tile index
#
proc IfuMap:DeHighlight { tile } {

    global IfuMap

    set canvas $IfuMap(canvas)

    $canvas itemconfig $tile -outline ""
    $canvas raise grid
}

#******************************************************************************
#
# Enable the edit buttons
#
proc IfuMap:EnableEditButtons { } {

    global IfuMap

    $IfuMap(path,undef_button) configure -state normal
    $IfuMap(path,sky_button) configure -state normal
    $IfuMap(path,prog_button1) configure -state normal
    $IfuMap(path,prog_button2) configure -state normal
    $IfuMap(path,prog_button3) configure -state normal
    $IfuMap(path,prog_button4) configure -state normal
    $IfuMap(path,dead_button) configure -state normal
}

#******************************************************************************
#
# Disable the edit buttons
#
proc IfuMap:DisableEditButtons { } {

    global IfuMap

    $IfuMap(path,undef_button) configure -state disabled
    $IfuMap(path,sky_button) configure -state disabled
    $IfuMap(path,prog_button1) configure -state disabled
    $IfuMap(path,prog_button2) configure -state disabled
    $IfuMap(path,prog_button3) configure -state disabled
    $IfuMap(path,prog_button4) configure -state disabled
    $IfuMap(path,dead_button) configure -state disabled
}

#******************************************************************************
#
# Enable the plot XY cut button
#
proc IfuMap:EnablePlotXYCutButton { } {

    global IfuMap

    $IfuMap(path,plotxycut_button) configure -state normal
}

#******************************************************************************
#
# Disable the plot XY cut button
#
proc IfuMap:DisablePlotXYCutButton { } {

    global IfuMap

    $IfuMap(path,plotxycut_button) configure -state disabled
}

#******************************************************************************
#
# Enable the save button
#
proc IfuMap:EnableSaveButton { } {

    global IfuMap

    $IfuMap(path,save_button) configure -state normal
}

#******************************************************************************
#
# Disable the save button
#
proc IfuMap:DisableSaveButton { } {

    global IfuMap

    $IfuMap(path,save_button) configure -state disabled
}

#******************************************************************************
#
# ???
#
# Arguments
#   type -
#
proc IfuMap:SetSelectionToType { type } {

    global IfuMapSelection

    set tile_list $IfuMapSelection
    foreach tile $tile_list {
        IfuMap:SetTileType $tile $type
    }

    # set tile types, types are obtained using GetTileType{}
    IfuMap:DisplayTypes

    # ???
    IfuMap:SetSelection ""

    # ???
    IfuMap:EnableSaveButton

    # ???
    IfuMap:UpdatePlotStates
}

#******************************************************************************
#
# Sets type for the given tile
#
# Arguments
#   tile - tile coordinates
#   type - type letter
#
proc IfuMap:SetTileType { tile type } {

    global IfuMapTileTypes

    set IfuMapTileTypes($tile) $type
}

#******************************************************************************
#
# Gets type for the given tile
#
# Arguments
#   tile - tile coordinates
#
# Returns type letter
#
proc IfuMap:GetTileType { tile } {

    global IfuMapTileTypes

    return $IfuMapTileTypes($tile)
}

#******************************************************************************
#
# Sets colour, value and type for all grid tiles
#
# Origin of types ???
#
# Arguments
#   colourAname -
#   valuesAname -
#
proc IfuMap:ConfigTiles { colourAname valuesAname } {

    global IfuMap
    global IfuMapTileValues

    upvar $colourAname colourA
    upvar $valuesAname valuesA

    set nrows $IfuMap(nrows)
    set ncols $IfuMap(ncols)
    
    for {set i 1} {$i<=$nrows} {incr i 1} {
        for {set j 1} {$j<=$ncols} {incr j 1} { 
            set tile ${i}:${j}

            set value $valuesA($i,$j)
            set IfuMapTileValues($tile) $value

#            set type $typesA($i,$j)
#            IfuMap:SetTileType $tile $type
        }
    }

    # set tile colours
    IfuMap:ColourTiles colourA

    # set tile types, types are obtained using GetTileType{}
    IfuMap:DisplayTypes

    # ???
    IfuMap:UpdatePlotStates     
}

#******************************************************************************
#
# ???
#
proc IfuMap:UpdatePlotStates { } {

    global IfuMap

    set type_list "P1 P2 P3 P4"

    foreach type $type_list {set non_empty($type) 0}
    for {set i 1} {$i<=$IfuMap(nrows)} {incr i 1} {
        for {set j 1} {$j<=$IfuMap(ncols)} {incr j 1} {
            set tile ${i}:${j}
            set type [IfuMap:GetTileType $tile]
            foreach match_type $type_list {
                if {$type==$match_type} {
                    set non_empty($type) 1
                }
            }
        }
    }
    
    # enable or disable button according to type_list
    foreach type $type_list {
        if {$non_empty($type)} {
            $IfuMap(path,plot,$type) configure -state normal
        } else {
            $IfuMap(path,plot,$type) configure -state disabled
        }
    }
}

#******************************************************************************
#
# Colours each tile according to given scale data
#
# Arguments
#   colourAname - 2-dimension array containing scale data
#
proc IfuMap:ColourTiles { colourAname } {

    global IfuMap

    upvar $colourAname colourA
        
    set c $IfuMap(canvas)

    set nrows $IfuMap(nrows)
    set ncols $IfuMap(ncols)
    
    for {set i 1} {$i<=$nrows} {incr i 1} {
        for {set j 1} {$j<=$ncols} {incr j 1} { 
            set tile ${i}:${j}
            set scale $colourA($i,$j)
            set colour grey$scale
            $c itemconfig $tile -fill $colour -outline ""
        }
    }
}

#******************************************************************************
#
# Displays type in each tile.
#
# Types are obtained using GetTileType.
#
proc IfuMap:DisplayTypes { } {

    global IfuMap
    global IfuMapTileA
    global IfuMapTileTypes
    
    set canvas $IfuMap(canvas)

    set tilelst [array names IfuMapTileTypes]
    foreach tile $tilelst {
        set type [IfuMap:GetTileType $tile]
        $canvas itemconfig "${tile}:type" -text $type
    }
    $canvas raise text
    $canvas raise grid  
}

#******************************************************************************
#
# Creates the Help panel
#
proc IfuMap:HelpPanel { } {

    puts "Here 2"

    # put up "Help" panel
    toplevel .helpPanel
    wm title .helpPanel "Info for IfuMap Users"

    set mssg \
"The AAOmega-IFU instrument consists of a lenslet array feeding light to a 32\n\
by 16 array of fibres. The fibres are fed to the AAOmega spectrograph giving\n\
512 spectra.\n\
\n\
This tool takes as input a 2dfdr reduced AAOmega-IFU observation and displays\n\
the 512 IFU fibres in a 32 high by 16 wide grid of tiles. Fibres are assigned\n\
a grid member and type according to the XGPOS, YGPOS and TYPE columns in the\n\
input file's fibre table. An ASCII input file can also be used to set these.\n\
values.\n\
\n\
Pressing the 'Grid' checkbox toggles the lines between grid members on and\n\
off. Pressing the 'Text' checkbox togges the grid member type text on and\n\
off. Moving the mouse to an individual grid tile causes the status line to\n\
show the tile's position, type and ??? value ???.\n\
\n\
Pressing the mouse button in an individual grid tile selects the tile.\n\
Holding down the mouse button while dragging the mouse over several tiles\n\
selects all tiles in the mouse path. Holding down the Control key while\n\
pressing the mouse button in an individual grid tile adds the tile to the\n\
currently selected set.\n\
\n\
Double-clicking in an individual grid tile produces a plot of corresponding\n\
fibre's spectrum.\n\
\n\
Pressing the 'Sky', 'Unused', 'Prog1', 'Prog2' 'Prog3' or 'Prog4' button\n\
sets the selected tiles (and the corresponding fibres) to that type. The\n\
designation can be removed using the 'Undo' button.\n\
\n\
Individual or sets of Program fibres can have their contents plotted using\n\
the 'Plot P1', 'Plot P2', 'Plot P3' or 'Plot P4' button.
\n\
Once fibres have their types established the information can be written to\n\
an ASCII text file. The file can then be used with 2dfdr to re-reduce the\n
observation.\n\
\n\
Pressing the 'Grayscale...' button brings up the Grey Scale Setting and\n\
statistics panel. This panel is used to ???\n\
\n\
Ability to plot X,Y cuts ???\n\
\n\
The format for the input ASCII text file should be as follows. Blank lines\n\
or lines beginning with '#' (comments) are ignored. Other lines should have\n\
the following 3 columns:\n\
    1. X position in the grid, 1-16\n\
    2. Y position in the grid, 1-32\n\
    3. Type, either F- guide, N- broken, P- program (science),\n\
       S- sky or U- unallocated, unused\n\
"
    puts "Here 1"

    text .helpPanel.tx -text $mssg -state disable -yscroll ".scroll set" \
        -wrap none -yscrollcommand "tx.v_scroll set" \
        -justify left -anchor c -relief raised -bd 1
    scrollbar .tx.v_scroll -command ".tx.text yview"
    pack .tx.v_scroll -side right -fill y
    pack .helpPanel.tx -anchor c -ipadx 5m -ipady 5m

    frame .helpPanel.f -relief raised -bd 1
    pack .helpPanel.f -fill both -expand yes
    button .helpPanel.f.close -text "Close" -command "destroy .helpPanel"
    pack .helpPanel.f.close -pady 3m

    grab set .helpPanel       ;# user can only interact with this window
    tkwait window .helpPanel  ;# returns when window is destroyed
}

#******************************************************************************
#
# Read fibre info from ASCII text file
#
# The file contents are used to set fibre types. Types are set using 
# proc SetTileType.
#
# Arguments
#   file       - name of file to read
#
proc IfuMap:ReadAsciiFile { file } {

    global IfuMap
        
    # if filename was not passed, put up panel to obtain it from user
    if {$file==""} {
        set file_types {{ "IFU input text file" { "*.dat" }}}
        set file [tk_getOpenFile -filetypes $file_types \
                        -title "IFU Text Input File"]
    }

    # quit if user chose cancel
    if {$file==""} {
        return
    }

    # initialise return values to empty strings
    for {set i 1} {$i<=$IfuMap(nrows)} {incr i 1} {
        for {set j 1} {$j<=$IfuMap(ncols)} {incr j 1} {
            set tile ${i}:${j}
            IfuMap:SetTileType $tile ""
       }
    }   
    
    # open, read into $data and close file
    set fid [open $file r]
    set data [read $fid]
    close $fid

    # process $data lines, one line each time around this loop
    set elem_list ""
    set lines [split $data \n]
    foreach line $lines {

        # remove leading and trailing blanks, ignore blank and comments lines
        set line [string trim $line " "]
        if {$line==""} { continue }
        if {[string range $line 0 0]=="#"} { continue }

        # grap X and Y positions, and Type from line
        set x_idx [lindex $line 0]
        set y_idx [lindex $line 1]
        set type  [lindex $line 2]

        # check Type
        # must be F-guide, N-broken, dead or no fibre,
        # P-program (science), S-sky or U-unallocated, unused
        if {$type!="F" && $type!="N"
            && $type!="P" && $type!="S" && $type!="U"} {
            tk_messageBox -type ok -icon error \
                -message "Invalid Type \"$type\" Found"
            return
        }

        # set type
        puts "$y_idx,$x_idx type=$type"
        set tile ${y_idx}:${x_idx}
        IfuMap:SetTileType $tile $type
    }
}

#******************************************************************************
#
# Write fibre info to ASCII text file
#
proc IfuMap:WriteAsciiFile { } {
        
    global IfuMap

    # get filename
    set filename $IfuMap(filename,fibfile)

    # Set default values to be empty strings
    for {set i 1} {$i<=$IfuMap(nrows)} {incr i 1} {
        for {set j 1} {$j<=$IfuMap(ncols)} {incr j 1} {
            set tile ${i}:${j}
            set typesA($i,$j) [IfuMap:GetTileType $tile]
        }
    }   
        
    # start with header lines
    set lines ""
    set line "# X       Y       Type"
    lappend lines $line
    
    # Set default values to be empty strings
    for {set i 1} {$i<=$IfuMap(nrows)} {incr i 1} {
        for {set j 1} {$j<=$IfuMap(ncols)} {incr j 1} {
            if  {$typesA($i,$j)!=""} {
                set type [string range $typesA($i,$j) 0 0]
                set type_id [string range $typesA($i,$j) 1 1]
            }
        }
    }
    set data [join $lines \n]   
        
    # open file, write data and close file
    set fid [open $filename w]
    puts $fid $data
    close $fid

    # Output data into IFU file
    set ifu_lst ""
    for {set yid 1} {$yid<=$IfuMap(nrows)} {incr yid 1} {
        for {set xid 1} {$xid<=$IfuMap(ncols)} {incr xid 1} {
            lappend ifu_lst $IfuMap(data,$yid,$xid)
        }
    }

    if {[file exists $ifu_fitsfile]} {
# THIS DOESN'T LOOK RIGHT, Mike, SO I CHANGED IT rwh 1May07
#       file delete file
        file delete $ifu_fitsfile
    }

    set fob [fits open $ifu_fitsfile 2]
    $fob insert image -p -32 2 "16 32"
    $fob put image 1 $ifu_lst
    $fob close

    # disable save button
    IfuMap:DisableSaveButton
}

#******************************************************************************
#
# ??? How is this connected with ReadAsciiFile, WriteAsciiFile ???
#
# Arguments
#   fname      -
#   dataAname  -
#   sdataAname -
#
# Returns ???
#
proc IfuMap:GetIfuData { fname dataAname sdataAname } {

    upvar $dataAname dataA
    upvar $sdataAname sdataA

    # open file for reading
    set fob [fits open $fname 0]

    # get number of rows and columns
    set imgdim [$fob info imgdim]
    set ncols [lindex $imgdim 0]
    set nrows [lindex $imgdim 1]

    # get data as list
    set lst [$fob get image]

    # close file
    $fob close
        
    set dataA(list) $lst
    set dataA(sorted_list) [lsort -real $lst]
        
    # define min and max to use for scaling from data list
    set minmax [IfuMap:SelectMinMax $lst]
    set min [lindex $minmax 0]
    set max [lindex $minmax 1]
    
    # define data Array from FITS image data
    set dataA(ncols) $ncols
    set dataA(nrows) $nrows
    set i 1; set j 1
    foreach itm $lst {
        set dataA($i,$j) $itm
        incr j 1
        if {$j>$ncols} {
            incr i 1
            set j 1
        }
    }
        
    # create scaled version of data using min and max values
    IfuMap:GenScaleData $min $max dataA sdataA
    
    return [list $min $max]
}

#******************************************************************************
#
# Gets minimum and maximum values from given data
#
# Clipping of negatives and some extreme values is done before taking the
# minimum and maximum.
#
# Arguments
#   lst - data list
#
# Returns ???
#
proc IfuMap:SelectMinMax { lst } {
    
    # set $lst2 to sorted data
    set lst2 [lsort -real $lst]
    
    # set $lst3 to sorted positives
    set pidx [vexpr SUM(lst2<=0.0)]
    set lst3 [lrange $lst2 $pidx end]

    # min,max is first,last of sorted data
    set min [lindex $lst3 0]
    set max [lindex $lst3 end]

    # return if number of positives < 10
    set n [llength $lst3]
    if {$n<10} {
        return [list $min $max]
    }
        
    # look for extreme positive points that can be clipped
    set idx_90 [expr int (0.9*$n)]
    set val [lindex $lst3 [expr $idx_90-1]]
    set low_range [expr $val-$min]
    set last_idx [expr $n-2]
#    puts "min=$min, max=$max, idx_90=$idx_90, val=$val"
#    puts "low_range=$low_range, last_idx=$last_idx"
    set extreme_found 0
    for {set idx $idx_90} {$idx<=$last_idx} {incr idx} {
        set val [lindex $lst3 $idx]
        set range [expr $val-$min]
        set ratio [expr $low_range/$range]
        set pc [expr int(100.0*$idx/$n)]
        #MNB#  puts "$idx ($pc%)-> $val $min : $range $ratio"
        if {$ratio<0.1} {
            set extreme_found 1
            set extreme_idx [expr $idx -1]
            #MNB#  puts "Extreme found [lindex $lst3 $extreme_idx]"
            #MNB#  puts "Extreme found [lindex $lst3 [expr $extreme_idx +1]]"
            #MNB#  puts "Extreme found [lindex $lst3 [expr $extreme_idx+2]]"
            break
        }
    } 

    if {$extreme_found} {
        return [list $min [lindex $lst3 $extreme_idx]]
    } else {
        return [list $min $max]
    }
}

#******************************************************************************
#
# Normalises data between min and max to between 0 and 100
#
# Arguments
#   min        - input minimum
#   max        - input maximum
#   dataAname  - input data
#   sdataAname - returned scaled data
#
proc IfuMap:GenScaleData { min max dataAname sdataAname } {

    upvar $dataAname dataA
    upvar $sdataAname sdataA
        
    # scaled data array has same dimensions as input data array
    set nrows $dataA(nrows)
    set ncols $dataA(ncols)
    set sdataA(nrows) $nrows
    set sdataA(ncols) $ncols

    set range [expr $max-$min]
    for {set i 1} {$i<=$nrows} {incr i 1} {
        for {set j 1} {$j<=$ncols} {incr j 1} {
            set val $dataA($i,$j)
            set scale [expr int(100.0*($val-$min)/$range)]
            if {$scale>100} {
                set scale 100
            }
            if {$scale<0} {
                set scale 0
            }
            set sdataA($i,$j) $scale
        }
    }

    # set ???
    set sdataA(used_min) $min
    set sdataA(used_max) $max
}

#******************************************************************************
#
# ???
#
# Arguments
#   lst3 - 
#
# Returns ???
#
proc IfuMap:AutoSelectFindExtreme { lst3 } {

    # Check the supplied (assumed sorted) data list to see
    # if the max point is extreme. We use 
    set min [lindex $lst3 0]
    set max [lindex $lst3 end]
    set last_idx [expr $n-1]
    set idx_90 [expr int (0.9*$n)]
    
    for {set idx1 $idx_90} {$idx1<$last_idx} {incr idx1 1} {
        set val [lindex $lst3 $idx1]
        set low_range [expr $val-$min]
        for {set idx1 [expr $idx1+1]} {$idx2<=$last_idx} {incr idx2 1} {
            set val [lindex $lst3 $idx2]
            set high_range [expr $val-$min]
            set ratio [expr 1.0*$low_range/$high_range]
            if {$ratio<0.1} {
                set extreme_flag 1
            }
        }
    }
        
    set last_range [expr $max-$min]
    set extreme_found 0
    for {set idx $last_idx} {$idx>=$idx_90} {incr idx -1} {
        set val [lindex $lst3 $idx]
        set range [expr $val-$min]
        set ratio [expr $range/$last_range]
        if {$ratio<0.1} {
            set extreme_found 1
            set extreme_idx [expr $idx+1]
            break
        }
    } 
        
    return $extreme_found
}

#******************************************************************************
#
# Reads reduced input FITS file ???
#
# Writes FITS file ifufile ???
#
# Reads XGPOS, YGPOS and TYPE column values from the file fibre table.
#
# Types are set using SetTileType
#
# Arguments
#   redfile    - input file name
#   dataAname  - returned data
#
proc IfuMap:ReadRedFile { redfile dataAname } {

    global IfuMap
    global IfuMap_tableA

    upvar $dataAname dataA

    # read fibre table from file
    # sets tableA(n_cols) to column count, tableA(n_rows) to row count
    # tableA(i,j) to table data, columns to column names
    set columns [FitsOps:TableDump $redfile *FIBRES_IFU tableA]

    # get column number for XGPOS,YGPOS,TYPE
    set colx [lsearch $columns XGPOS]
    incr colx 1
    set coly [lsearch $columns YGPOS]
    incr coly 1
    set coltype [lsearch $columns TYPE]
    incr coltype 1

    # set data array dimensions
    set dataA(nrows) $IfuMap(nrows)
    set dataA(ncols) $IfuMap(ncols)

    # open file for reading
    set fob [fits open $redfile 0]
    
    # process next fibre spectrum each time around this loop
    # !!! NULL DATA VALUES ARE NOT PROPERLY HANDLED HERE !!!
    for {set row 1} {$row<=$tableA(n_rows)} {incr row 1} {
                
        # map fibre number to XGPOS,YGPOS
        set xid $tableA($row,$colx)        ;# XGPOS value
        set yid $tableA($row,$coly)        ;# YGPOS value
#        set IfuMap(row,$yid,$xid) $row
        set IfuMap_tableA($yid,$xid) $row

        # set $row_ptr to list -> pointer dataType numElements
        set row_ptr [$fob load irows $row $row]

        # convert pointer to Tcl list of flux values
        # sets row_lst to list -> dataList dataType naxesList
        set row_lst [eval "ptr2lst $row_ptr"]

        # use vexpr to average fibre flux values, save result in IfuMap(data
        set sum [vexpr "sum(row_lst)/nelem(row_lst)"]
        set dataA($yid,$xid) $sum
#        set IfuMap(data,$yid,$xid) $sum

        # get fibre type
        set tile ${yid}:${xid} 
        IfuMap:SetTileType $tile $tableA($row,$coltype)

        # get sdata
        set sdataA($yid,$xid) 50
    }
        
    $fob close
        
    # Output the data into IFU file
#    set ifu_lst ""
#    for {set yid 1} {$yid<=$IfuMap(nrows)} {incr yid 1} {
#        for {set xid 1} {$xid<=$IfuMap(ncols)} {incr xid 1} {
#            lappend ifu_lst $IfuMap(data,$yid,$xid)
#        }
#    }

    # delete $ifufile if it exists
#    if {[file exists $ifufile]} {
#        file delete $ifufile
#    }

#    # write $ifufile ???
#    set fob [fits open $ifufile 2]
#    $fob insert image -p -32 2 "16 32"
#    $fob put image 1 $ifu_lst
#    $fob close
}

#******************************************************************************
#
# Converts given tile index to corresponding fibre number
#
# Arguments
#   tile - tile index
#
# Returns fibre number
#
proc IfuMap:Tile2FibreNo { tile } {

    global IfuMap_tableA

    set rowcol [split $tile :]
    set row [lindex $rowcol 0]
    set col [lindex $rowcol 1]
    set fibreno $IfuMap_tableA($row,$col)
    return $fibreno
}

#******************************************************************************
#
# Creates plot of ???
#
# Arguments
#   fitsfile -
#   rowlist  - list of fibre numbers to include in plot
#
proc IfuMap:ViewSumRows { fitsfile rowlist } {

    set tmpfile tmp_sumrows.dat
    IfuMap:SumRows $fitsfile $rowlist $tmpfile
    tclGraph:OpenGraph $tmpfile
}

#******************************************************************************
#
# ???
#
# Arguments
#   fitsfile - input file name
#   rowlist  - list of fibre number to include
#   outfile  - optional
#
# Returns ???
#
proc IfuMap:SumRows { fitsfile rowlist { outfile "" } } {

    set waveflag 1

    # open file for reading
    set fob [fits open $fitsfile 0]
        
    set n [llength $rowlist]
    set first 1
    foreach row $rowlist {
        set row_ptr [$fob load irows $row $row]

        # Convert pointer to Tcl list
        set row_lst [eval "ptr2lst $row_ptr"]
        if {$first} {
            set first 0
            set sum_lst $row_lst
        } else {
            set sum_lst [vexpr sum_lst+row_lst]
        }
    }
    if {$n>1} {
        set sum_lst [vexpr sum_lst/$n]
    }
    $fob close
    if {$outfile==""} {
        return $sum_lst
    }
    
    # Output tcl list to text file
    if {[file exists $outfile]} {
        file delete $outfile
    }
    set fid [open $outfile w]
    
    if {$waveflag} {
        set crval1 [FitsOps:GetKeyWord $fitsfile CRVAL1]
        set cdelt1 [FitsOps:GetKeyWord $fitsfile CDELT1]
        set crpix1 [FitsOps:GetKeyWord $fitsfile CRPIX1]
        set ctype1 [FitsOps:GetKeyWord $fitsfile CTYPE1]
        set cunit1 [FitsOps:GetKeyWord $fitsfile CUNIT1]
        set idx 0
        foreach itm $sum_lst {
            incr idx 1
            set xval [expr $crval1+($idx-$crpix1)*$cdelt1]
            puts $fid "$xval $itm"
        }
        close $fid
        return
    }
        
    set idx 1
    foreach itm $sum_lst {
        puts $fid "$idx $itm"
        incr idx 1
    }
    close $fid
}

#******************************************************************************
#
# Creates panel to obtain reduced and fibre info files from user
#
# Arguments
#   redfile - optional
#   fibfile - optional
#
# Returns ???
#
proc IfuMap:GetFilesPanel { { redfile "" } { fibfile "" } } {

    global IfuMapGetFilesPanel
        
    set root .ifuMapGetFilesPanel
    destroy $root
    toplevel $root
    wm transient $root .    ;# this keeps it on top
    wm title $root "IFU Map Files:"
        
    frame $root.top -relief raised -bd 2
    frame $root.bot -relief raised -bd 2
    pack $root.top $root.bot -fill x
    
    set lroot $root.top

    # label, entry and browse button to get reduced file
    label $lroot.l1 -text "AAOmegaIFU Reduced file:"
    entry $lroot.e1 -width 32 -textvariable IfuMapGetFilesPanel(disvar,redfile)
    button $lroot.b1 -text "Browse..." -command {
        set file_types {{"Reduced file" {"*red.fits"}}}
        set redfile [tk_getOpenFile -filetypes $file_types]
        set redfile [file tail $redfile]
        set IfuMapGetFilesPanel(disvar,redfile) $redfile
    }

    # label, entry and browse button to get fibre info file
    label $lroot.l2 -text "Fibre Info File:"
    entry $lroot.e2 -width 32 -textvariable IfuMapGetFilesPanel(disvar,fibfile)
    button $lroot.b2 -text "Browse..." -command {
        set file_types {{"IFU Fibre data file" {"*.dat"}}}
        set redfile [tk_getOpenFile -filetypes $file_types]
        set redfile [file tail $redfile]
        set IfuMapGetFilesPanel(disvar,fibfile) $redfile
    }

    grid $lroot.l1 -row 1 -column 1 -sticky e
    grid $lroot.e1 -row 1 -column 2 
    grid $lroot.b1 -row 1 -column 3 -sticky w
    grid $lroot.l2 -row 2 -column 1 -sticky e
    grid $lroot.e2 -row 2 -column 2 
    grid $lroot.b2 -row 2 -column 3 -sticky w
    
    button $root.bot.ok -text OK -width 7 \
        -command "set IfuMapGetFilesPanel(flag,ok) 1; destroy $root"
    button $root.bot.ca -text Cancel -width 7 -command "destroy $root"
    pack  $root.bot.ok $root.bot.ca -side left -padx 4m -pady 4m -expand 1
    
    set IfuMapGetFilesPanel(flag,ok) 0
    set IfuMapGetFilesPanel(disvar,redfile) $redfile
    set IfuMapGetFilesPanel(disvar,fibfile) $fibfile
    tkwait window $root
    if {!$IfuMapGetFilesPanel(flag,ok)} {
        return ""
    }
    set redfile $IfuMapGetFilesPanel(disvar,redfile)
    set fibfile $IfuMapGetFilesPanel(disvar,fibfile)
    if {$redfile==""} {
        tk_messageBox -message "Must select a reduced file!"
        return ""
    }
    if {$fibfile==""} {
        tk_messageBox -message "Must specify a fibre info file!"
        return ""
    }
    if {![file exists $redfile]} {
        tk_messageBox -message "Cannot find reduced file $redfile"
        return ""
    }
    if {![file exists $fibfile]} {
        set reply [tk_messageBox -type yesno \
                       -message "$fibfile does not exist. Create new one?"]
        if {$reply=="no"} {
            return ""
        }
    }
        
    return [list $redfile $fibfile]
    IfuMap $redfile $fibfile
}

#******************************************************************************
#
# X and Y Cut plots
#
proc IfuMap:PlotXYCut { } {

    global IfuMapSelection

    set selected_tiles IfuMapSelection
    set tile [lindex $selected_tiles]
    set rowcol [split  $tile :]
    set i [lindex $rowcol 0]
    set j [lindex $rowcol 1]
    IfuMap:XYCuts $i $j
}

#******************************************************************************
#
# ???
#
# Arguments
#   i0 - 
#   j0 - 
#
proc IfuMap:XYCuts { i0 j0 } {
    
    global IfuMap
    global IfuMapTileValues

    set xfile tmp_xcut.dat
    set yfile tmp_ycut.dat
    set xcut_data ""
    set ycut_data ""

    set nrows $IfuMap(nrows)
    set ncols $IfuMap(ncols)
    
    for {set i 1} {$i<=$nrows} {incr i 1} {
       set tile ${i}:${j0} 
       set value $IfuMapTileValues($tile)
       lappend ycut_data [list [expr $i-0.5] $value]
       lappend ycut_data [list [expr $i+0.5] $value]
    }
    
    for {set j 1} {$j<=$ncols} {incr j 1} {
       set tile ${i0}:${j} 
       set value $IfuMapTileValues($tile)
       lappend xcut_data [list [expr $j-0.5] $value]
       lappend xcut_data [list [expr $j+0.5] $value]
    }
    
    set fid [open $xfile w]
    puts $fid [join $xcut_data \n]
    close $fid
    tclGraph:OpenGraph $xfile

    set fid [open $yfile w]
    puts $fid [join $ycut_data \n]
    close $fid
    tclGraph:OpenGraph $yfile
}

#******************************************************************************
#
# Test code
#
proc IfuMap:Test { } {

    global tcl_interactive
    global IfuMap

    # reduced file we are looking at
    set redfile 23jun10040red.fits

    # temp ifu file we generate
    set tmpifu_fitsfile tmp_ifutest.fits

    # fibinfo file that we edit
    set tst_fibinfofile fibinfo_test.dat

    IfuMap $redfile $tst_fibinfofile $tmpifu_fitsfile

    # If this is not interactive session
    # then wait until IfuMap root window is destroyed
    if {$tcl_interactive==0} {
        tkwait window $IfuMap(root)
    }
}
